from sempy_labs._helper_functions import (
    _base_api,
    resolve_workspace_id,
    resolve_item_id,
)
from typing import Optional
import pandas as pd
from uuid import UUID
from sempy._utils._log import log
import pyarrow as pa


@log
def execute_query(
    dataflow: str | UUID,
    query_name: str,
    custom_mashup_document: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
) -> pd.DataFrame:
    """
    Executes a query against a dataflow and returns the result.

    This is a wrapper function for the following API: `Query Execution - Execute Query <https://learn.microsoft.com/rest/api/fabric/dataflow/query-execution/execute-query>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    dataflow : str | uuid.UUID
        Name or ID of the dataflow.
    query_name : str
        The name of the query to execute from the dataflow (or from the custom mashup document if provided).
    custom_mashup_document : str, default=None
        Optional custom mashup document to override the dataflow's default mashup.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the results of the query execution.
    """

    workspace_id = resolve_workspace_id(workspace)
    item_id = resolve_item_id(item=dataflow, type="Dataflow", workspace=workspace_id)

    payload = {
        "queryName": query_name,
    }

    if custom_mashup_document:
        payload["customMashupDocument"] = custom_mashup_document

    response = _base_api(
        request=f"/v1/workspaces/{workspace_id}/dataflows/{item_id}/executeQuery",
        method="post",
        payload=payload,
        lro_return_status_code=True,
        status_codes=[200, 202],
    )

    # with pa.ipc.open_stream(response.content) as reader:
    #    df = reader.read_pandas()

    with pa.ipc.open_stream(response.content) as reader:
        batches = [batch for batch in reader]

    table = pa.Table.from_batches(batches)

    df = table.to_pandas(
        split_blocks=True,
        self_destruct=True,
    )

    return df



from ._items import (
    get_workspace_consumption_rules,
    get_background_operation_rules,
    set_workspace_consumption_rules,
    set_background_operation_rules,
    delete_workspace_consumption_rules,
    delete_background_operation_rules,
)

__all__ = [
    "get_workspace_consumption_rules",
    "get_background_operation_rules",
    "set_workspace_consumption_rules",
    "set_background_operation_rules",
    "delete_workspace_consumption_rules",
    "delete_background_operation_rules",
]



from uuid import UUID
from typing import List
from sempy_labs._helper_functions import (
    _base_api,
)
from sempy._utils._log import log


@log
def generate_embed_token(dataset_ids: List[UUID], report_ids: List[UUID]):

    if not isinstance(dataset_ids, list):
        dataset_ids = [dataset_ids]
    if not isinstance(report_ids, list):
        report_ids = [report_ids]

    payload = {
        "datasets": [{"id": dataset_id} for dataset_id in dataset_ids],
        "reports": [{"id": report_id, "allowEdit": True} for report_id in report_ids],
    }

    response = _base_api(
        request="/v1.0/myorg/GenerateToken",
        method="post",
        client="fabric_sp",
        payload=payload,
    )
    return response.json().get("token")



import sempy.fabric as fabric
import pandas as pd
import sempy_labs._icons as icons
from typing import Optional, List
import base64
import requests
from sempy._utils._log import log
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    resolve_workspace_id,
    _decode_b64,
    _base_api,
    resolve_item_id,
    create_item,
    _create_dataframe,
)
from sempy.fabric.exceptions import FabricHTTPException
from os import PathLike
from uuid import UUID
import os

_notebook_prefix = "notebook-content."


def get_notebook_definition_base(
    notebook_name: str,
    workspace: Optional[str | UUID] = None,
    format: Optional[str] = None,
) -> pd.DataFrame:

    workspace_id = resolve_workspace_id(workspace)
    item_id = resolve_item_id(item=notebook_name, type="Notebook", workspace=workspace)

    url = f"v1/workspaces/{workspace_id}/notebooks/{item_id}/getDefinition"
    if format == "ipynb":
        url += f"?format={format}"

    result = _base_api(
        request=url,
        method="post",
        lro_return_json=True,
        status_codes=None,
    )

    return pd.json_normalize(result["definition"]["parts"])


def get_notebook_type(
    notebook_name: str, workspace: Optional[str | UUID] = None
) -> str:

    df_items = get_notebook_definition_base(
        notebook_name=notebook_name, workspace=workspace
    )

    file_path = df_items[df_items["path"].str.startswith(_notebook_prefix)][
        "path"
    ].iloc[0]

    _, file_extension = os.path.splitext(file_path)

    return file_extension[1:]


@log
def get_notebook_definition(
    notebook_name: str,
    workspace: Optional[str | UUID] = None,
    decode: bool = True,
    format: Optional[str] = None,
) -> str:
    """
    Obtains the notebook definition.

    This is a wrapper function for the following API: `Items - Get Notebook Definition <https://learn.microsoft.com/rest/api/fabric/notebook/items/get-notebook-definition>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    notebook_name : str
        The name of the notebook.
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    decode : bool, default=True
        If True, decodes the notebook definition file into .ipynb format.
        If False, obtains the notebook definition file in base64 format.
    format : str, default=None
        The only supported value is ipynb
        If provided the format will be in standard .ipynb otherwise the format will be in source code format which is GIT friendly ipynb

    Returns
    -------
    str
        The notebook definition.
    """

    df_items = get_notebook_definition_base(
        notebook_name=notebook_name, workspace=workspace, format=format
    )
    df_items_filt = df_items[df_items["path"].str.startswith(_notebook_prefix)]
    payload = df_items_filt["payload"].iloc[0]

    if decode:
        result = _decode_b64(payload)
    else:
        result = payload

    return result


@log
def import_notebook_from_web(
    notebook_name: str,
    url: str,
    description: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
    overwrite: bool = False,
    folder: Optional[str | PathLike] = None,
):
    """
    Creates a new notebook within a workspace based on a Jupyter notebook hosted in the web.

    Note: When specifying a notebook from GitHub, please use the raw file path. Note that if the non-raw file path is specified, the url will be
    converted to the raw URL as the raw URL is needed to obtain the notebook content.

    Parameters
    ----------
    notebook_name : str
        The name of the notebook to be created.
    url : str
        The url of the Jupyter Notebook (.ipynb)
    description : str, default=None
        The description of the notebook.
        Defaults to None which does not place a description.
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    overwrite : bool, default=False
        If set to True, overwrites the existing notebook in the workspace if it exists.
    folder : str | os.PathLike, default=None
        The folder within the workspace where the notebook will be created.
        Defaults to None which places the notebook in the root of the workspace.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    # Fix links to go to the raw github file
    starting_text = "https://github.com/"
    starting_text_len = len(starting_text)
    if url.startswith(starting_text):
        url = f"https://raw.githubusercontent.com/{url[starting_text_len:]}".replace(
            "/blob/", "/"
        )

    response = requests.get(url)
    if response.status_code != 200:
        raise FabricHTTPException(response)

    dfI = fabric.list_items(workspace=workspace, type="Notebook")
    dfI_filt = dfI[dfI["Display Name"] == notebook_name]
    if len(dfI_filt) == 0:
        create_notebook(
            name=notebook_name,
            notebook_content=response.content.decode("utf-8"),
            workspace=workspace_id,
            description=description,
            format="ipynb",
            folder=folder,
        )
    elif len(dfI_filt) > 0 and overwrite:
        print(f"{icons.info} Overwrite of notebooks is currently not supported.")
        # update_notebook_definition(
        #    name=notebook_name, notebook_content=response.content, workspace=workspace
        # )
    else:
        raise ValueError(
            f"{icons.red_dot} The '{notebook_name}' already exists within the '{workspace_name}' workspace and 'overwrite' is set to False."
        )


@log
def create_notebook(
    name: str,
    notebook_content: str,
    type: str = "py",
    description: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
    format: Optional[str] = None,
    folder: Optional[str | PathLike] = None,
):
    """
    Creates a new notebook with a definition within a workspace.

    This is a wrapper function for the following API: `Items - Create Notebook Definition <https://learn.microsoft.com/rest/api/fabric/notebook/items/create-notebook>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    name : str
        The name of the notebook to be created.
    notebook_content : str
        The Jupyter notebook content (not in Base64 format).
    type : str, default="py"
        The notebook type.
    description : str, default=None
        The description of the notebook.
        Defaults to None which does not place a description.
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    format : str, default=None
        If 'ipynb' is provided than notebook_content should be standard ipynb format
        otherwise notebook_content should be GIT friendly format
    folder : str | os.PathLike, default=None
        The folder within the workspace where the notebook will be created.
        Defaults to None which places the notebook in the root of the workspace.
    """

    notebook_payload = base64.b64encode(notebook_content.encode("utf-8")).decode(
        "utf-8"
    )

    definition_payload = {
        "parts": [
            {
                "path": f"{_notebook_prefix}{type}",
                "payload": notebook_payload,
                "payloadType": "InlineBase64",
            }
        ],
    }

    if format == "ipynb":
        definition_payload["format"] = "ipynb"

    create_item(
        name=name,
        type="Notebook",
        workspace=workspace,
        description=description,
        definition=definition_payload,
        folder=folder,
    )


@log
def update_notebook_definition(
    name: str,
    notebook_content: str,
    workspace: Optional[str | UUID] = None,
    format: Optional[str] = None,
):
    """
    Updates an existing notebook with a new definition.

    This is a wrapper function for the following API: `Items - Update Notebook Definition <https://learn.microsoft.com/rest/api/fabric/notebook/items/update-notebook-definition>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    name : str
        The name of the notebook to be updated.
    notebook_content : str
        The Jupyter notebook content (not in Base64 format).
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    format : str, default=None
        If 'ipynb' is provided than notebook_content should be standard ipynb format
        otherwise notebook_content should be GIT friendly format
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    notebook_payload = base64.b64encode(notebook_content.encode("utf-8")).decode(
        "utf-8"
    )
    item_id = resolve_item_id(item=name, type="Notebook", workspace=workspace)
    type = get_notebook_type(notebook_name=name, workspace=workspace)

    payload = {
        "definition": {
            "parts": [
                {
                    "path": f"{_notebook_prefix}{type}",
                    "payload": notebook_payload,
                    "payloadType": "InlineBase64",
                }
            ],
        },
    }

    if format == "ipynb":
        payload["definition"]["format"] = "ipynb"

    _base_api(
        request=f"v1/workspaces/{workspace_id}/notebooks/{item_id}/updateDefinition",
        payload=payload,
        method="post",
        lro_return_status_code=True,
        status_codes=None,
    )

    print(
        f"{icons.green_dot} The '{name}' notebook was updated within the '{workspace_name}' workspace."
    )


@log
def list_notebooks(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows the notebooks within a workspace.

    This is a wrapper function for the following API: `Items - List Notebooks <https://learn.microsoft.com/rest/api/fabric/notebook/items/list-notebooks>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the notebooks within a workspace.
    """

    columns = {
        "Notebook Id": "string",
        "Notebook Name": "string",
        "Description": "string",
    }
    df = _create_dataframe(columns=columns)

    workspace_id = resolve_workspace_id(workspace)

    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/notebooks", uses_pagination=True
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            rows.append(
                {
                    "Notebook Id": v.get("id"),
                    "Notebook Name": v.get("displayName"),
                    "Description": v.get("description"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def search_notebooks(
    search_string: str,
    notebook: Optional[str | UUID] = None,
    workspace: Optional[str | UUID | List[str | UUID]] = None,
) -> pd.DataFrame:
    """
    Searches notebooks within a workspace or across multiple workspaces for a given search string.

    Parameters
    ----------
    search_string : str
        The string to search for within the notebook definitions.
    notebook : str | uuid.UUID, default=None
        The name or ID of a specific notebook to search within.
        Defaults to None which searches across all notebooks in the specified workspace(s).
    workspace : str | uuid.UUID | list, default=None
        The name or ID of the workspace or a list of workspaces to search within.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
        If a list is provided, it should contain workspace names or IDs.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the notebooks that contain the search string in their definitions.
        The dataframe includes the workspace name, workspace ID, notebook name, and notebook ID.
    """

    if not workspace:
        workspace_id = resolve_workspace_id(workspace)
        workspace_ids = [workspace_id]
    elif isinstance(workspace, str):
        workspace_id = resolve_workspace_id(workspace)
        workspace_ids = [workspace_id]
    elif isinstance(workspace, list):
        workspace_ids = [resolve_workspace_id(ws) for ws in workspace]
    else:
        raise ValueError(
            "Workspace must be a string, UUID, or a list of strings/UUIDs."
        )

    dfW = fabric.list_workspaces()
    dfW_filt = dfW[dfW["Id"].isin(workspace_ids)]

    columns = {
        "Workspace Name": "string",
        "Workspace Id": "string",
        "Notebook Name": "string",
        "Notebook Id": "string",
    }
    df = _create_dataframe(columns=columns)

    rows = []
    for _, r in dfW_filt.iterrows():
        w_id = r["Id"]
        w_name = r["Name"]
        dfN = list_notebooks(workspace=w_id)
        if notebook is not None:
            item_id = resolve_item_id(item=notebook, type="Notebook", workspace=w_id)
            dfN = dfN[dfN["Notebook Id"] == item_id]
        for _, n in dfN.iterrows():
            notebook_id = n["Notebook Id"]
            notebook_name = n["Notebook Name"]
            definition = _base_api(
                request=f"v1/workspaces/{w_id}/notebooks/{notebook_id}/getDefinition",
                method="post",
                client="fabric_sp",
                status_codes=None,
                lro_return_json=True,
            )
            for part in definition.get("definition").get("parts", []):
                payload = _decode_b64(part.get("payload", ""))
                if part.get("path") in ("notebook-content.py", "notebook-content.sql"):
                    if search_string in payload:
                        rows.append(
                            {
                                "Workspace Name": w_name,
                                "Workspace Id": w_id,
                                "Notebook Name": notebook_name,
                                "Notebook Id": notebook_id,
                            }
                        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df



from ._save_report import (
    save_report_as_pbip,
)
from ._reportwrapper import (
    ReportWrapper,
    connect_report,
)
from ._paginated import (
    get_report_datasources,
)
from ._generate_report import (
    create_report_from_reportjson,
    get_report_definition,
    update_report_from_reportjson,
    create_model_bpa_report,
)
from ._download_report import download_report
from ._report_functions import (
    get_report_json,
    # report_dependency_tree,
    clone_report,
    launch_report,
    # translate_report_titles
)
from ._report_rebind import (
    report_rebind,
    report_rebind_all,
)
from ._report_bpa_rules import report_bpa_rules
from ._report_bpa import run_report_bpa
from ._export_report import (
    export_report,
)
from ._endorsement import (
    set_endorsement,
)
from ._upgrade_to_pbir import (
    upgrade_to_pbir,
)
from ._items import (
    list_reports,
    get_report,
)

__all__ = [
    "create_report_from_reportjson",
    "update_report_from_reportjson",
    "get_report_json",
    # report_dependency_tree,
    "export_report",
    "clone_report",
    "launch_report",
    # translate_report_titles,
    "report_rebind",
    "report_rebind_all",
    "get_report_definition",
    "create_model_bpa_report",
    "ReportWrapper",
    "report_bpa_rules",
    "run_report_bpa",
    "get_report_datasources",
    "download_report",
    "save_report_as_pbip",
    "connect_report",
    "set_endorsement",
    "upgrade_to_pbir",
    "list_reports",
    "get_report",
]



import pandas as pd
from uuid import UUID
from sempy_labs._helper_functions import (
    resolve_capacity_id,
    _create_dataframe,
    _update_dataframe_datatypes,
    _base_api,
)
import sempy_labs._icons as icons
from sempy._utils._log import log


def _surge_api(capacity, url, payload, method="get", status_code=200, return_json=True):

    capacity_id = resolve_capacity_id(capacity)

    response = _base_api(
        request=f"capacities/{capacity_id}/{url}",
        client="internal",
        method=method,
        payload=payload,
        status_codes=status_code,
    )

    if return_json:
        return response.json()
    else:
        return None  # response


@log
def get_workspace_consumption_rules(
    capacity: str | UUID = None, return_dataframe: bool = True
) -> pd.DataFrame | dict:
    """
    Retrieves the workspace consumption surge protection rules for the specified capacity.

    Workspace Consumption
        When CU consumption by a single workspace reaches the rejection threshold, reject new operation requests and block the workspace for the specified amount of time.

    Parameters
    ----------
    capacity : str | uuid.UUID, default=None
        The capacity name or ID.
        Defaults to None which resolves to the capacity id of the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the capacity name of the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame | dict
        A pandas dataframe showing the workspace consumption surge protection rules for the specified capacity,
        or a dictionary if return_dataframe is set to False.
    """

    response_json = _surge_api(capacity=capacity, url="detectionRules", payload=None)

    if not return_dataframe:
        return response_json
    else:
        columns = {
            "Usage Threshold": "float",
            "Blocked Duration": "str",
            "Blocked Duration Policy": "str",
            "Detection Rule Id": "str",
            "Detection Rule Type": "str",
        }
        df = _create_dataframe(columns=columns)
        rows = []
        workspace_action = response_json.get("blockWorkspaceAction", {})
        rows.append(
            {
                "Usage Threshold": response_json.get(
                    "highWorkspaceUsageCondition", {}
                ).get("usageThreshold"),
                "Blocked Duration": workspace_action.get("blockedDuration"),
                "Blocked Duration Policy": workspace_action.get(
                    "blockedDurationPolicy"
                ),
                "Detection Rule Id": response_json.get("detectionRuleId"),
                "Detection Rule Type": response_json.get("detectionRuleType"),
            }
        )
        if rows:
            df = pd.DataFrame(rows)
            _update_dataframe_datatypes(dataframe=df, column_map=columns)
        return df


@log
def get_background_operation_rules(
    capacity: str | UUID = None, return_dataframe: bool = True
) -> pd.DataFrame | dict:
    """
    Retrieves the background operation surge protection rules for the specified capacity.

    Background Operations
        When total CU consumption reaches the rejection threshold, reject new background operation requests. When total CU consumption drops below the recovery threshold, accept new background operation requests.

    Parameters
    ----------
    capacity : str | uuid.UUID, default=None
        The capacity name or ID.
        Defaults to None which resolves to the capacity id of the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the capacity name of the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame | dict
        A pandas dataframe showing the background operation surge protection rules for the specified capacity,
        or a dictionary if return_dataframe is set to False.
    """

    response_json = _surge_api(
        capacity=capacity, url="surgeProtectionRules", payload=None
    )

    if not return_dataframe:
        return response_json
    else:
        columns = {
            "Rule Instance Id": "str",
            "Rule Type Id": "int",
            "Rule Id": "str",
            "Rejection Threshold Percentage": "int",
            "Recovery Threshold Percentage": "int",
            "Utilization Type": "str",
            "Throttling Level": "str",
        }

        df = _create_dataframe(columns=columns)
        rows = []
        for rule in response_json.get("rules", []):
            trig = rule.get("triggeredThrottlingLevel", {})
            throt = rule.get("usageThrottlingCriteria", {})
            rows.append(
                {
                    "Rule Instance Id": rule.get("ruleInstanceId"),
                    "Rule Type Id": rule.get("ruleTypeId"),
                    "Rule Id": rule.get("ruleId"),
                    "Rejection Threshold Percentage": throt.get(
                        "triggerThresholdPercentage"
                    ),
                    "Recovery Threshold Percentage": throt.get(
                        "recoveryThresholdPercentage"
                    ),
                    "Utilization Type": trig.get("utilizationType"),
                    "Throttling Level": trig.get("throttlingLevel"),
                }
            )
        if rows:
            df = pd.DataFrame(rows)
            _update_dataframe_datatypes(dataframe=df, column_map=columns)
        return df


@log
def set_workspace_consumption_rules(
    capacity: str | UUID = None,
    rejection_threshold: int = 75,
    block_duration_hours: int = 24,
    block_indefinitely: bool = False,
) -> dict:
    """
    Sets the workspace consumption surge protection rules for the specified capacity.

    Workspace Consumption
        When total CU consumption by a single workspace reaches the rejection threshold, reject new operation requests and block the workspace for the specified amount of time.

    Parameters
    ----------
    capacity : str | uuid.UUID, default=None
        The capacity name or ID.
        Defaults to None which resolves to the capacity id of the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the capacity name of the workspace of the notebook.
    rejection_threshold : int, default=75
        The CU consumption percentage threshold (1-100) at which new operation requests will be rejected.
    block_duration_hours : int, default=24
        The duration in hours to block the workspace when the rejection threshold is reached. Must be at least 1 hour.
        Ignored if block_indefinitely is set to True.
    block_indefinitely : bool, default=False
        If True, the workspace will be blocked indefinitely when the rejection threshold is reached.

    Returns
    -------
    dict
        A dictionary showing the workspace consumption surge protection rules for the specified capacity.
    """

    if not block_indefinitely and block_duration_hours < 1:
        raise ValueError(
            f"{icons.red_dot} The block_duration_hours must be at least 1 hour."
        )

    if rejection_threshold < 1 or rejection_threshold > 100:
        raise ValueError(
            f"{icons.red_dot} The rejection_threshold must be between 1 and 100."
        )

    payload = {
        "detectionRuleId": None,
        "detectionRuleType": "detectHighWorkspaceUsageAndBlockWorkspace",
        "blockWorkspaceAction": {
            "blockedDurationPolicy": (
                "indefinite" if block_indefinitely else "fixedDuration"
            ),
            "blockedDuration": (
                None if block_indefinitely else f"PT{block_duration_hours}H"
            ),
        },
        "highWorkspaceUsageCondition": {
            "usageThreshold": rejection_threshold,
        },
    }

    return _surge_api(
        capacity=capacity, url="detectionRules", payload=payload, method="post"
    )


@log
def set_background_operation_rules(
    capacity: str | UUID = None,
    rejection_threshold: int = 75,
    recovery_threshold: int = 25,
) -> dict:
    """
    Sets the background operation surge protection rules for the specified capacity.

    Background Operations
        When total CU consumption reaches the rejection threshold, reject new background operation requests. When total CU consumption drops below the recovery threshold, accept new background operation requests.

    Parameters
    ----------
    capacity : str | uuid.UUID, default=None
        The capacity name or ID.
        Defaults to None which resolves to the capacity id of the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the capacity name of the workspace of the notebook.
    rejection_threshold : int, default=75
        The CU consumption percentage threshold (1-100) at which new background operation requests will be rejected.
    recovery_threshold : int, default=25
        The CU consumption percentage threshold (5-100) at which new background operation requests will be accepted.

    Returns
    -------
    dict
        A dictionary showing the background operation surge protection rules for the specified capacity.
    """

    if rejection_threshold < 1 or rejection_threshold > 100:
        raise ValueError(
            f"{icons.red_dot} The rejection_threshold must be between 1 and 100."
        )
    if recovery_threshold < 5 or recovery_threshold > 100:
        raise ValueError(
            f"{icons.red_dot} The recovery_threshold must be between 5 and 100."
        )
    if recovery_threshold >= rejection_threshold:
        raise ValueError(
            f"{icons.red_dot} The recovery_threshold must be less than the rejection_threshold."
        )

    payload = {
        "rules": [
            {
                "ruleTypeId": 1,
                "ruleInstanceId": 1,
                "triggeredThrottlingLevel": {
                    "utilizationType": "Background",
                    "throttlingLevel": "Extreme",
                },
                "usageThrottlingCriteria": {
                    "triggerThresholdPercentage": rejection_threshold,
                    "recoveryThresholdPercentage": recovery_threshold,
                },
            }
        ]
    }

    _surge_api(
        capacity=capacity,
        url="surgeProtectionRules",
        payload=payload,
        method="put",
        return_json=False,
    )

    return get_background_operation_rules(capacity=capacity, return_dataframe=False)


@log
def delete_workspace_consumption_rules(capacity: str | UUID = None):
    """
    Deletes the workspace consumption surge protection rules for the specified capacity.

    Workspace Consumption
        When total CU consumption by a single workspace reaches the rejection threshold, reject new operation requests and block the workspace for the specified amount of time.

    Parameters
    ----------
    capacity : str | uuid.UUID, default=None
        The capacity name or ID.
        Defaults to None which resolves to the capacity id of the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the capacity name of the workspace of the notebook.
    """

    rules = get_workspace_consumption_rules(capacity=capacity, return_dataframe=False)
    if not rules:
        print(f"{icons.yellow_dot} No workspace consumption rules found to delete.")
        return

    for v in rules.get("value", []):
        rule_id = v.get("detectionRuleId")
        break

    _surge_api(
        capacity=capacity,
        url=f"detectionRules/{rule_id}",
        payload=None,
        method="delete",
        status_code=204,
        return_json=False,
    )

    print(f"{icons.green_dot} The workspace consumption rules deleted successfully.")


@log
def delete_background_operation_rules(capacity: str | UUID = None):
    """
    Deletes the background operation surge protection rules for the specified capacity.

    Background Operations
        When total CU consumption reaches the rejection threshold, reject new background operation requests. When total CU consumption drops below the recovery threshold, accept new background operation requests.

    Parameters
    ----------
    capacity : str | uuid.UUID, default=None
        The capacity name or ID.
        Defaults to None which resolves to the capacity id of the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the capacity name of the workspace of the notebook.
    """

    _surge_api(
        capacity=capacity,
        url="surgeProtectionRules",
        payload=None,
        method="delete",
        status_code=200,
        return_json=False,
    )

    print(
        f"{icons.green_dot} The background operation rules have been deleted successfully."
    )



import pandas as pd
from typing import Optional
from sempy_labs._helper_functions import (
    _base_api,
    _create_dataframe,
    delete_item,
    resolve_workspace_id,
)
from uuid import UUID
from sempy._utils._log import log


@log
def list_event_schema_sets(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Returns a list of Event Schema Sets from the specified workspace.

    This is a wrapper function for the following API: `Items - List Event Schema Sets <https://learn.microsoft.com/rest/api/fabric/eventschemaset/items/list-event-schema-sets>`_.
    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the Event Schema Sets within a workspace.
    """
    workspace_id = resolve_workspace_id(workspace)

    columns = {
        "Event Schema Set Name": "str",
        "Event Schema Set Id": "str",
        "Description": "str",
        "OneLake Root Path": "str",
    }
    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request=f"v1/workspaces/{workspace_id}/eventSchemaSets",
        client="fabric_sp",
        uses_pagination=True,
    )

    rows = []

    for r in responses:
        for v in r.get("value", []):
            row = {
                "Event Schema Set Name": v.get("displayName"),
                "Event Schema Set Id": v.get("id"),
                "Description": v.get("description"),
                "OneLake Root Path": v.get("properties", {}).get("oneLakeRootPath"),
            }
            rows.append(row)

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def delete_event_schema_set(
    event_schema_set: str | UUID, workspace: Optional[str | UUID] = None
):
    """
    Deletes an Event Schema Set from the specified workspace.

    This is a wrapper function for the following API: `Items - Delete Event Schema Set <https://learn.microsoft.com/rest/api/fabric/eventschemaset/items/delete-event-schema-set>`_.

    Parameters
    ----------
    event_schema_set : str | uuid.UUID
        The Event Schema Set name or ID to delete.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    delete_item(item=event_schema_set, type="EventSchemaSet", workspace=workspace)



from typing import Literal, Optional, List
from sempy_labs._helper_functions import (
    resolve_item_id,
    resolve_workspace_id,
    _base_api,
)
from uuid import UUID
from sempy._utils._log import log
import sempy_labs._icons as icons
from sempy_labs.kql_database import resolve_cluster_uri


@log
def nl_to_kql(
    kql_database: str | UUID,
    billing_item: str | UUID,
    billing_item_type: Literal["KQLQueryset", "KQLDashboard", "Eventhouse"],
    prompt: str,
    chat_messages: Optional[dict | List[dict]] = None,
    user_shots: Optional[List[dict]] = None,
    workspace: Optional[str | UUID] = None,
) -> str:
    """
    Returns a KQL query generated from natural language.

    This is a wrapper function for the following API: `Copilot - NL To KQL <https://learn.microsoft.com/rest/api/fabric/realtimeintelligence/copilot/nl-to-kql(beta)>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    kql_database : str | uuid.UUID
        The name or UUID of the KQL database.
    billing_item : str | uuid.UUID
        The name or ID of the item for the request. This can be a KQLQueryset, KQLDashboard, or Eventhouse item. This item is used for billing the request.
    billing_item_type : typing.Literal["KQLQueryset", "KQLDashboard", "Eventhouse"]
        The type of the billing item.
    prompt : str
        The natural language to generate the KQL query from.
    chat_messages : Optional[str | List[str]], default=None
        The `chat messages <https://learn.microsoft.com/rest/api/fabric/realtimeintelligence/copilot/nl-to-kql(beta)?tabs=HTTP#chatmessage>`_ for the request. The chat messages provide additional context for generating the KQL query if necessary.

        Example:
            chat_messages = [
                {
                    "content": "Content....",
                    "role": "User"
                },
                {
                    "content": "Content...",
                    "role": "Assistant"
                }
            ]
    user_shots : Optional[List[str]], default=None
        The `user shots <https://learn.microsoft.com/rest/api/fabric/realtimeintelligence/copilot/nl-to-kql(beta)?tabs=HTTP#usershot>`_ for the request. This consists of user provided pairs of natural language and KQL queries in order to help in generating the current requested KQL query.

        Example:
            user_shots = [
                {
                    "kqlQuery": "KQL Query....",
                    "naturalLanguage": "Natural language...."
                },
                {
                    "kqlQuery": "KQL Query...",
                    "naturalLanguage": "Natural language..."
                }
            ]
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    str
        The generated KQL query.
    """

    workspace_id = resolve_workspace_id(workspace=workspace)
    db_id = resolve_item_id(
        item=kql_database, type="KQLDatabase", workspace=workspace_id
    )
    cluster_uri = resolve_cluster_uri(kql_database=kql_database, workspace=workspace_id)

    valid_types = ["KQLQueryset", "KQLDashboard", "Eventhouse"]
    if billing_item_type not in valid_types:
        raise ValueError(
            f"{icons.red_dot} The billing_item_type must be either 'KQLQueryset', 'KQLDashboard', or 'Eventhouse'."
        )

    billing_item_id = resolve_item_id(
        item=billing_item, type=billing_item_type, workspace=workspace_id
    )

    payload = {
        "clusterUrl": cluster_uri,
        "databaseName": db_id,
        "itemIdForBilling": billing_item_id,
        "naturalLanguage": prompt,
    }

    if chat_messages:
        if isinstance(chat_messages, str):
            chat_messages = [chat_messages]
        payload["chatMessages"] = chat_messages
    if user_shots:
        if isinstance(user_shots, str):
            user_shots = [user_shots]
        payload["userShots"] = user_shots

    response = _base_api(
        request=f"/v1/workspaces/{workspace_id}/realtimeintelligence/nltokql?beta=True",
        payload=payload,
        method="post",
        client="fabric_sp",
    )
    return response.json().get("kqlQuery")



from ._copilot import (
    approved_for_copilot,
    set_endorsement,
    make_discoverable,
)
from ._caching import (
    enable_query_caching,
)


__all__ = [
    "approved_for_copilot",
    "set_endorsement",
    "make_discoverable",
    "enable_query_caching",
]



import sempy
import pandas as pd
from typing import Optional
from uuid import UUID
from sempy._utils._log import log
import sempy_labs._icons as icons
from sempy_labs.lakehouse._get_lakehouse_columns import get_lakehouse_columns
from sempy_labs.tom import connect_semantic_model
from sempy_labs._helper_functions import (
    _convert_data_type,
    resolve_workspace_name_and_id,
    resolve_dataset_name_and_id,
)
from sempy_labs.directlake._sources import get_direct_lake_sources


@log
def direct_lake_schema_sync(
    dataset: str | UUID,
    workspace: Optional[str | UUID] = None,
    add_to_model: bool = False,
    remove_from_model: bool = False,
) -> pd.DataFrame:
    """
    Shows/adds columns which exist in the lakehouse but do not exist in the semantic model (only for tables in the semantic model).

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    add_to_model : bool, default=False
        If set to True, columns which exist in the lakehouse but do not exist in the semantic model are added to the semantic model. No new tables are added.
    remove_from_model : bool, default=False
        If set to True, columns which exist in the semantic model but do not exist in the lakehouse are removed from the semantic model. No new tables are removed.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the status of columns in the semantic model and lakehouse (prior to adding/removing them from the model using this function).
    """

    sempy.fabric._client._utils._init_analysis_services()
    import Microsoft.AnalysisServices.Tabular as TOM

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    sources = get_direct_lake_sources(dataset=dataset_id, workspace=workspace_id)[0]
    if sources.get("itemType") != "Lakehouse":
        raise ValueError(
            f"{icons.red_dot} The source of the '{dataset_name}' semantic model within the '{workspace_name}' workspace is not a Fabric lakehouse. This function only supports Direct Lake semantic models which source from Fabric lakehouses."
        )

    lakehouse_id = sources.get("itemId")
    lakehouse_workspace_id = sources.get("workspaceId")

    lc = get_lakehouse_columns(lakehouse=lakehouse_id, workspace=lakehouse_workspace_id)

    readonly = True
    if add_to_model or remove_from_model:
        readonly = False
    df = pd.DataFrame(
        columns=[
            "TableName",
            "ColumnName",
            "SourceTableName",
            "SourceColumnName",
            "Status",
        ]
    )

    rows_to_add = []
    rows_to_remove = []

    with connect_semantic_model(
        dataset=dataset_id, readonly=readonly, workspace=workspace_id
    ) as tom:
        # Check if the columns in the semantic model exist in the lakehouse
        for c in tom.all_columns():
            column_name = c.Name
            table_name = c.Parent.Name
            partition_name = next(p.Name for p in c.Table.Partitions)
            p = c.Table.Partitions[partition_name]
            if p.Mode == TOM.ModeType.DirectLake:
                entity_name = p.Source.EntityName
                source_column = c.SourceColumn
                lc_filt = lc[
                    (lc["Table Name"] == entity_name)
                    & (lc["Column Name"] == source_column)
                ]
                # Remove column from model if it doesn't exist in the lakehouse
                if lc_filt.empty:
                    rows_to_remove.append(
                        {
                            "TableName": table_name,
                            "ColumnName": column_name,
                            "SourceTableName": entity_name,
                            "SourceColumnName": source_column,
                            "Status": "Not in lakehouse",
                        }
                    )
                    if remove_from_model:
                        tom.remove_object(object=c)
                        print(
                            f"{icons.green_dot} The '{table_name}'[{column_name}] column has been removed from the '{dataset_name}' semantic model within the '{workspace_name}' workspace."
                        )

        # Check if the lakehouse columns exist in the semantic model
        for _, r in lc.iterrows():
            source_table = r["Table Name"]
            source_column = r["Column Name"]
            dType = r["Data Type"]

            if any(
                p.Source.EntityName == source_table
                for p in tom.all_partitions()
                if p.Mode == TOM.ModeType.DirectLake
            ):
                table_name = next(
                    t.Name
                    for t in tom.model.Tables
                    for p in t.Partitions
                    if p.Mode == TOM.ModeType.DirectLake
                    and p.Source.EntityName == source_table
                )

                if not any(
                    c.SourceColumn == source_column and c.Parent.Name == table_name
                    for c in tom.all_columns()
                ):
                    rows_to_add.append(
                        {
                            "TableName": table_name,
                            "ColumnName": None,
                            "SourceTableName": source_table,
                            "SourceColumnName": source_column,
                            "Status": "Not in semantic model",
                        }
                    )

                    if add_to_model:
                        dt = _convert_data_type(dType)
                        tom.add_data_column(
                            table_name=table_name,
                            column_name=source_column,
                            source_column=source_column,
                            data_type=dt,
                        )
                        print(
                            f"{icons.green_dot} The '{source_column}' column in the '{source_table}' lakehouse table was added to the '{dataset_name}' semantic model within the '{workspace_name}' workspace."
                        )

        COLUMNS = [
            "TableName",
            "ColumnName",
            "SourceTableName",
            "SourceColumnName",
            "Status",
        ]
        df = pd.DataFrame(rows_to_add + rows_to_remove, columns=COLUMNS)
        return df



import sempy.fabric as fabric
from sempy_labs._helper_functions import (
    resolve_workspace_id,
    resolve_workspace_name_and_id,
    create_relationship_name,
    format_dax_object_name,
    resolve_dataset_name_and_id,
    _update_dataframe_datatypes,
    _base_api,
    _create_dataframe,
    _run_spark_sql_query,
)
from sempy._utils._log import log
import pandas as pd
from typing import Optional
import sempy_labs._icons as icons
from uuid import UUID
import json
from collections import defaultdict


@log
def get_object_level_security(
    dataset: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Shows the object level security for the semantic model.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the object level security for the semantic model.
    """

    from sempy_labs.tom import connect_semantic_model

    columns = {
        "Role Name": "string",
        "Object Type": "string",
        "Table Name": "string",
        "Object Name": "string",
        "Metadata Permission": "string",
    }
    df = _create_dataframe(columns=columns)

    rows = []

    with connect_semantic_model(
        dataset=dataset, readonly=True, workspace=workspace
    ) as tom:

        for r in tom.model.Roles:
            for tp in r.TablePermissions:
                for cp in tp.ColumnPermissions:
                    rows.append(
                        {
                            "Role Name": r.Name,
                            "Object Type": "Column",
                            "Table Name": tp.Name,
                            "Object Name": cp.Name,
                            "Metadata Permission": cp.Permission,
                        }
                    )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
    return df


@log
def list_tables(
    dataset: str | UUID, workspace: Optional[str | UUID] = None, extended: bool = False
) -> pd.DataFrame:
    """
    Shows a semantic model's tables and their properties.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    extended : bool, default=False
        Adds additional columns including Vertipaq statistics.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the semantic model's tables and their properties.
    """

    from sempy_labs.tom import connect_semantic_model

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    columns = {
        "Name": "string",
        "Description": "string",
        "Hidden": "bool",
        "Data Category": "string",
        "Type": "string",
        "Refresh Policy": "bool",
        "Source Expression": "string",
    }

    df = _create_dataframe(columns=columns)

    with connect_semantic_model(
        dataset=dataset_id, workspace=workspace_id, readonly=True
    ) as tom:
        if extended:
            dict_df = fabric.evaluate_dax(
                dataset=dataset_id,
                workspace=workspace_id,
                dax_string="""
                EVALUATE SELECTCOLUMNS(FILTER(INFO.STORAGETABLECOLUMNS(), [COLUMN_TYPE] = "BASIC_DATA"),[DIMENSION_NAME],[DICTIONARY_SIZE])
                """,
            )
            dict_sum = dict_df.groupby("[DIMENSION_NAME]")["[DICTIONARY_SIZE]"].sum()
            data = fabric.evaluate_dax(
                dataset=dataset_id,
                workspace=workspace_id,
                dax_string="""EVALUATE SELECTCOLUMNS(INFO.STORAGETABLECOLUMNSEGMENTS(),[TABLE_ID],[DIMENSION_NAME],[USED_SIZE])""",
            )
            data_sum = (
                data[
                    ~data["[TABLE_ID]"].str.startswith("R$")
                    & ~data["[TABLE_ID]"].str.startswith("U$")
                    & ~data["[TABLE_ID]"].str.startswith("H$")
                ]
                .groupby("[DIMENSION_NAME]")["[USED_SIZE]"]
                .sum()
            )
            hier_sum = (
                data[data["[TABLE_ID]"].str.startswith("H$")]
                .groupby("[DIMENSION_NAME]")["[USED_SIZE]"]
                .sum()
            )
            rel_sum = (
                data[data["[TABLE_ID]"].str.startswith("R$")]
                .groupby("[DIMENSION_NAME]")["[USED_SIZE]"]
                .sum()
            )
            uh_sum = (
                data[data["[TABLE_ID]"].str.startswith("U$")]
                .groupby("[DIMENSION_NAME]")["[USED_SIZE]"]
                .sum()
            )
            rc = fabric.evaluate_dax(
                dataset=dataset_id,
                workspace=workspace_id,
                dax_string="""
                SELECT [DIMENSION_NAME],[ROWS_COUNT] FROM $SYSTEM.DISCOVER_STORAGE_TABLES
                WHERE RIGHT ( LEFT ( TABLE_ID, 2 ), 1 ) <> '$'
            """,
            )

            model_size = (
                dict_sum.sum()
                + data_sum.sum()
                + hier_sum.sum()
                + rel_sum.sum()
                + uh_sum.sum()
            )

        rows = []
        for t in tom.model.Tables:
            t_name = t.Name
            t_type = (
                "Calculation Group"
                if t.CalculationGroup
                else (
                    "Calculated Table"
                    if tom.is_calculated_table(table_name=t.Name)
                    else "Table"
                )
            )
            ref = bool(t.RefreshPolicy)
            ref_se = t.RefreshPolicy.SourceExpression if ref else None

            new_data = {
                "Name": t_name,
                "Description": t.Description,
                "Hidden": t.IsHidden,
                "Data Category": t.DataCategory,
                "Type": t_type,
                "Refresh Policy": ref,
                "Source Expression": ref_se,
            }

            if extended:
                dict_size = dict_sum.get(t_name, 0)
                data_size = data_sum.get(t_name, 0)
                h_size = hier_sum.get(t_name, 0)
                r_size = rel_sum.get(t_name, 0)
                u_size = uh_sum.get(t_name, 0)
                total_size = data_size + dict_size + h_size + r_size + u_size

                new_data.update(
                    {
                        "Row Count": (
                            rc[rc["DIMENSION_NAME"] == t_name]["ROWS_COUNT"].iloc[0]
                            if not rc.empty
                            else 0
                        ),
                        "Total Size": total_size,
                        "Dictionary Size": dict_size,
                        "Data Size": data_size,
                        "Hierarchy Size": h_size,
                        "Relationship Size": r_size,
                        "User Hierarchy Size": u_size,
                        "Partitions": int(len(t.Partitions)),
                        "Columns": sum(
                            1 for c in t.Columns if str(c.Type) != "RowNumber"
                        ),
                        "% DB": (
                            round((total_size / model_size) * 100, 2)
                            if model_size not in (0, None, float("nan"))
                            else 0.0
                        ),
                    }
                )

            rows.append(new_data)

        df = pd.DataFrame(rows)

        if extended:
            column_map = {
                "Row Count": "int",
                "Total Size": "int",
                "Dictionary Size": "int",
                "Data Size": "int",
                "Hierarchy Size": "int",
                "Relationship Size": "int",
                "User Hierarchy Size": "int",
                "Partitions": "int",
                "Columns": "int",
                "% DB": "float",
            }

            _update_dataframe_datatypes(dataframe=df, column_map=column_map)

    return df


@log
def list_annotations(
    dataset: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Shows a semantic model's annotations and their properties.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the semantic model's annotations and their properties.
    """

    from sempy_labs.tom import connect_semantic_model

    workspace_id = resolve_workspace_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    columns = {
        "Object Name": "string",
        "Parent Object Name": "string",
        "Object Type": "string",
        "Annotation Name": "string",
        "Annotation Value": "string",
    }
    df = _create_dataframe(columns=columns)

    rows = []
    with connect_semantic_model(
        dataset=dataset_id, readonly=True, workspace=workspace_id
    ) as tom:

        mName = tom.model.Name
        for a in tom.model.Annotations:
            objectType = "Model"
            aName = a.Name
            aValue = a.Value
            rows.append(
                {
                    "Object Name": mName,
                    "Parent Object Name": None,
                    "Object Type": objectType,
                    "Annotation Name": aName,
                    "Annotation Value": aValue,
                }
            )
        for t in tom.model.Tables:
            objectType = "Table"
            tName = t.Name
            for ta in t.Annotations:
                taName = ta.Name
                taValue = ta.Value
                rows.append(
                    {
                        "Object Name": tName,
                        "Parent Object Name": mName,
                        "Object Type": objectType,
                        "Annotation Name": taName,
                        "Annotation Value": taValue,
                    }
                )
            for p in t.Partitions:
                pName = p.Name
                objectType = "Partition"
                for pa in p.Annotations:
                    paName = pa.Name
                    paValue = pa.Value
                    rows.append(
                        {
                            "Object Name": pName,
                            "Parent Object Name": tName,
                            "Object Type": objectType,
                            "Annotation Name": paName,
                            "Annotation Value": paValue,
                        }
                    )
            for c in t.Columns:
                objectType = "Column"
                cName = c.Name
                for ca in c.Annotations:
                    caName = ca.Name
                    caValue = ca.Value
                    rows.append(
                        {
                            "Object Name": cName,
                            "Parent Object Name": tName,
                            "Object Type": objectType,
                            "Annotation Name": caName,
                            "Annotation Value": caValue,
                        }
                    )
            for ms in t.Measures:
                objectType = "Measure"
                measName = ms.Name
                for ma in ms.Annotations:
                    maName = ma.Name
                    maValue = ma.Value
                    rows.append(
                        {
                            "Object Name": measName,
                            "Parent Object Name": tName,
                            "Object Type": objectType,
                            "Annotation Name": maName,
                            "Annotation Value": maValue,
                        }
                    )
            for h in t.Hierarchies:
                objectType = "Hierarchy"
                hName = h.Name
                for ha in h.Annotations:
                    haName = ha.Name
                    haValue = ha.Value
                    rows.append(
                        {
                            "Object Name": hName,
                            "Parent Object Name": tName,
                            "Object Type": objectType,
                            "Annotation Name": haName,
                            "Annotation Value": haValue,
                        }
                    )
        for d in tom.model.DataSources:
            dName = d.Name
            objectType = "Data Source"
            for da in d.Annotations:
                daName = da.Name
                daValue = da.Value
                rows.append(
                    {
                        "Object Name": dName,
                        "Parent Object Name": mName,
                        "Object Type": objectType,
                        "Annotation Name": daName,
                        "Annotation Value": daValue,
                    }
                )
        for r in tom.model.Relationships:
            rName = r.Name
            objectType = "Relationship"
            for ra in r.Annotations:
                raName = ra.Name
                raValue = ra.Value
                rows.append(
                    {
                        "Object Name": rName,
                        "Parent Object Name": mName,
                        "Object Type": objectType,
                        "Annotation Name": raName,
                        "Annotation Value": raValue,
                    }
                )
        for cul in tom.model.Cultures:
            culName = cul.Name
            objectType = "Translation"
            for cula in cul.Annotations:
                culaName = cula.Name
                culaValue = cula.Value
                rows.append(
                    {
                        "Object Name": culName,
                        "Parent Object Name": mName,
                        "Object Type": objectType,
                        "Annotation Name": culaName,
                        "Annotation Value": culaValue,
                    }
                )
        for e in tom.model.Expressions:
            eName = e.Name
            objectType = "Expression"
            for ea in e.Annotations:
                eaName = ea.Name
                eaValue = ea.Value
                rows.append(
                    {
                        "Object Name": eName,
                        "Parent Object Name": mName,
                        "Object Type": objectType,
                        "Annotation Name": eaName,
                        "Annotation Value": eaValue,
                    }
                )
        for per in tom.model.Perspectives:
            perName = per.Name
            objectType = "Perspective"
            for pera in per.Annotations:
                peraName = pera.Name
                peraValue = pera.Value
                rows.append(
                    {
                        "Object Name": perName,
                        "Parent Object Name": mName,
                        "Object Type": objectType,
                        "Annotation Name": peraName,
                        "Annotation Value": peraValue,
                    }
                )
        for rol in tom.model.Roles:
            rolName = rol.Name
            objectType = "Role"
            for rola in rol.Annotations:
                rolaName = rola.Name
                rolaValue = rola.Value
                rows.append(
                    {
                        "Object Name": rolName,
                        "Parent Object Name": mName,
                        "Object Type": objectType,
                        "Annotation Name": rolaName,
                        "Annotation Value": rolaValue,
                    }
                )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def list_columns(
    dataset: str | UUID,
    workspace: Optional[str | UUID] = None,
) -> pd.DataFrame:
    """
    Shows a semantic model's columns and their properties.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the semantic model's columns and their properties.
    """
    from sempy_labs.directlake._sources import (
        get_direct_lake_sources,
    )

    workspace_id = resolve_workspace_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    fabric.refresh_tom_cache(workspace=workspace)
    dfP = fabric.list_partitions(dataset=dataset_id, workspace=workspace_id)

    isDirectLake = any(r["Mode"] == "DirectLake" for i, r in dfP.iterrows())

    dfC = fabric.list_columns(dataset=dataset_id, workspace=workspace_id)

    if isDirectLake:
        dfC["Column Cardinality"] = None
        sql_statements = []
        sources = get_direct_lake_sources(dataset=dataset, workspace=workspace)[0]
        # lake_id = sources.get("itemId")
        lake_name = sources.get("itemName")
        # lake_workspace_id = sources.get("workspaceId")

        for table_name in dfC["Table Name"].unique():
            print(f"Gathering stats for table: '{table_name}'...")
            query = "SELECT "

            columns_in_table = dfC.loc[
                dfC["Table Name"] == table_name, "Column Name"
            ].unique()

            # Loop through columns within those tables
            for column_name in columns_in_table:
                scName = dfC.loc[
                    (dfC["Table Name"] == table_name)
                    & (dfC["Column Name"] == column_name),
                    "Source",
                ].iloc[0]
                lakeTName = dfC.loc[
                    (dfC["Table Name"] == table_name)
                    & (dfC["Column Name"] == column_name),
                    "Query",
                ].iloc[0]

                # Build the query to be executed dynamically
                query = f"{query}COUNT(DISTINCT({scName})) AS {scName}, "

            query = query[:-2]
            query = f"{query} FROM {lake_name}.{lakeTName}"
            sql_statements.append((table_name, query))

        for o in sql_statements:
            tName = o[0]
            query = o[1]

            # Run the query
            df = _run_spark_sql_query(query)

            for column in df.columns:
                x = df.collect()[0][column]
                for i, r in dfC.iterrows():
                    if r["Table Name"] == tName and r["Source"] == column:
                        dfC.at[i, "Column Cardinality"] = x

        # Remove column added temporarily
        dfC.drop(columns=["Query"], inplace=True)

    return dfC


@log
def list_lakehouses(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows the lakehouses within a workspace.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the lakehouses within a workspace.
    """

    columns = {
        "Lakehouse Name": "string",
        "Lakehouse ID": "string",
        "Description": "string",
        "OneLake Tables Path": "string",
        "OneLake Files Path": "string",
        "SQL Endpoint Connection String": "string",
        "SQL Endpoint ID": "string",
        "SQL Endpoint Provisioning Status": "string",
        "Schema Enabled": "bool",
        "Default Schema": "string",
        "Sensitivity Label Id": "string",
    }
    df = _create_dataframe(columns=columns)

    workspace_id = resolve_workspace_id(workspace)

    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/lakehouses",
        uses_pagination=True,
        client="fabric_sp",
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            prop = v.get("properties", {})
            sqlEPProp = prop.get("sqlEndpointProperties", {})
            default_schema = prop.get("defaultSchema", None)

            rows.append(
                {
                    "Lakehouse Name": v.get("displayName"),
                    "Lakehouse ID": v.get("id"),
                    "Description": v.get("description"),
                    "OneLake Tables Path": prop.get("oneLakeTablesPath"),
                    "OneLake Files Path": prop.get("oneLakeFilesPath"),
                    "SQL Endpoint Connection String": sqlEPProp.get("connectionString"),
                    "SQL Endpoint ID": sqlEPProp.get("id"),
                    "SQL Endpoint Provisioning Status": sqlEPProp.get(
                        "provisioningStatus"
                    ),
                    "Schema Enabled": True if default_schema else False,
                    "Default Schema": default_schema,
                    "Sensitivity Label Id": v.get("sensitivityLabel", {}).get(
                        "sensitivityLabelId"
                    ),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def list_datamarts(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows the datamarts within a workspace.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the datamarts within a workspace.
    """

    columns = {
        "Datamart Name": "string",
        "Datamart ID": "string",
        "Description": "string",
    }
    df = _create_dataframe(columns=columns)

    workspace_id = resolve_workspace_id(workspace)

    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/datamarts", uses_pagination=True
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            rows.append(
                {
                    "Datamart Name": v.get("displayName"),
                    "Datamart ID": v.get("id"),
                    "Description": v.get("description"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def update_item(
    item_type: str,
    current_name: str,
    new_name: str,
    description: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Updates the name/description of a Fabric item.

    Parameters
    ----------
    item_type: str
        Type of item to update.
    current_name : str
        The current name of the item.
    new_name : str
        The new name of the item.
    description : str, default=None
        A description of the item.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    item_type = item_type.replace(" ", "").capitalize()

    if item_type not in icons.itemTypes.keys():
        raise ValueError(
            f"{icons.red_dot} The '{item_type}' is not a valid item type. "
        )

    itemType = icons.itemTypes[item_type]

    dfI = fabric.list_items(workspace=workspace_id, type=item_type)
    dfI_filt = dfI[(dfI["Display Name"] == current_name)]

    if len(dfI_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{current_name}' {item_type} does not exist within the '{workspace_name}' workspace."
        )

    itemId = dfI_filt["Id"].iloc[0]

    payload = {"displayName": new_name}
    if description:
        payload["description"] = description

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/{itemType}/{itemId}",
        payload=payload,
        method="patch",
    )
    if description is None:
        print(
            f"{icons.green_dot} The '{current_name}' {item_type} within the '{workspace_name}' workspace has been updated to be named '{new_name}'"
        )
    else:
        print(
            f"{icons.green_dot} The '{current_name}' {item_type} within the '{workspace_name}' workspace has been updated to be named '{new_name}' and have a description of '{description}'"
        )


@log
def list_user_defined_functions(
    dataset: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Shows a list of the user-defined functions within a semantic model.

    Parameters
    ----------
    dataset: str | uuid.UUID
        Name or UUID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of the user-defined functions within a semantic model.
    """

    from sempy_labs.tom import connect_semantic_model

    columns = {
        "Function Name": "string",
        "Expression": "string",
        "Lineage Tag": "string",
    }
    df = _create_dataframe(columns=columns)
    rows = []
    with connect_semantic_model(dataset=dataset, workspace=workspace) as tom:
        for f in tom.model.Functions:
            rows.append(
                {
                    "Function Name": f.Name,
                    "Expression": f.Expression,
                    "Lineage Tag": f.LineageTag,
                }
            )

    if rows:
        df = pd.DataFrame(rows)

    return df


@log
def list_relationships(
    dataset: str | UUID, workspace: Optional[str | UUID] = None, extended: bool = False
) -> pd.DataFrame:
    """
    Shows a semantic model's relationships and their properties.

    Parameters
    ----------
    dataset: str | uuid.UUID
        Name or UUID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    extended : bool, default=False
        Fetches extended column information.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the object level security for the semantic model.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    fabric.refresh_tom_cache(workspace=workspace)

    dfR = fabric.list_relationships(dataset=dataset_id, workspace=workspace_id)
    dfR["From Object"] = format_dax_object_name(dfR["From Table"], dfR["From Column"])
    dfR["To Object"] = format_dax_object_name(dfR["To Table"], dfR["To Column"])

    if extended:
        # Used to map the Relationship IDs
        rel = fabric.evaluate_dax(
            dataset=dataset_id,
            workspace=workspace_id,
            dax_string="""
                SELECT
                [ID] AS [RelationshipID]
                ,[Name]
                FROM $SYSTEM.TMSCHEMA_RELATIONSHIPS
                """,
        )

        # USED_SIZE shows the Relationship Size where TABLE_ID starts with R$
        cs = fabric.evaluate_dax(
            dataset=dataset_id,
            workspace=workspace_id,
            dax_string="""
                SELECT
                [TABLE_ID]
                ,[USED_SIZE]
                FROM $SYSTEM.DISCOVER_STORAGE_TABLE_COLUMN_SEGMENTS
                """,
        )

        def parse_value(text):
            ind = text.rfind("(") + 1
            output = text[ind:]
            output = output[:-1]
            return output

        cs["RelationshipID"] = cs["TABLE_ID"].apply(parse_value).astype("uint64")
        relcs = pd.merge(
            cs[["RelationshipID", "TABLE_ID", "USED_SIZE"]],
            rel,
            on="RelationshipID",
            how="left",
        )

        dfR["Used Size"] = None
        for i, r in dfR.iterrows():
            relName = r["Relationship Name"]

            filtered_cs = relcs[
                (relcs["Name"] == relName) & (relcs["TABLE_ID"].str.startswith("R$"))
            ]
            sumval = filtered_cs["USED_SIZE"].sum()
            dfR.at[i, "Used Size"] = sumval

        column_map = {
            "Used Size": "int",
        }

        _update_dataframe_datatypes(dataframe=dfR, column_map=column_map)

    return dfR


@log
def list_kpis(
    dataset: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Shows a semantic model's KPIs and their properties.

    Parameters
    ----------
    dataset: str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the KPIs for the semantic model.
    """

    from sempy_labs.tom import connect_semantic_model

    workspace_id = resolve_workspace_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    columns = {
        "Table Name": "string",
        "Measure Name": "string",
        "Target Expression": "string",
        "Target Format String": "string",
        "Target Description": "string",
        "Status Expression": "string",
        "Status Graphic": "string",
        "Status Description": "string",
        "Trend Expression": "string",
        "Trend Graphic": "string",
        "Trend Description": "string",
    }
    df = _create_dataframe(columns=columns)

    rows = []
    with connect_semantic_model(
        dataset=dataset_id, workspace=workspace_id, readonly=True
    ) as tom:

        for t in tom.model.Tables:
            for m in t.Measures:
                if m.KPI is not None:
                    rows.append(
                        {
                            "Table Name": t.Name,
                            "Measure Name": m.Name,
                            "Target Expression": m.KPI.TargetExpression,
                            "Target Format String": m.KPI.TargetFormatString,
                            "Target Description": m.KPI.TargetDescription,
                            "Status Graphic": m.KPI.StatusGraphic,
                            "Status Expression": m.KPI.StatusExpression,
                            "Status Description": m.KPI.StatusDescription,
                            "Trend Expression": m.KPI.TrendExpression,
                            "Trend Graphic": m.KPI.TrendGraphic,
                            "Trend Description": m.KPI.TrendDescription,
                        }
                    )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def list_semantic_model_objects(
    dataset: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Shows a list of semantic model objects.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.


    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of objects in the semantic model
    """
    from sempy_labs.tom import connect_semantic_model

    columns = {
        "Parent Name": "string",
        "Object Name": "string",
        "Object Type": "string",
    }
    df = _create_dataframe(columns=columns)

    rows = []
    with connect_semantic_model(
        dataset=dataset, workspace=workspace, readonly=True
    ) as tom:
        for t in tom.model.Tables:
            if t.CalculationGroup is not None:
                rows.append(
                    {
                        "Parent Name": t.Parent.Name,
                        "Object Name": t.Name,
                        "Object Type": "Calculation Group",
                    }
                )

                for ci in t.CalculationGroup.CalculationItems:
                    rows.append(
                        {
                            "Parent Name": t.Name,
                            "Object Name": ci.Name,
                            "Object Type": str(ci.ObjectType),
                        }
                    )
            elif any(str(p.SourceType) == "Calculated" for p in t.Partitions):
                rows.append(
                    {
                        "Parent Name": t.Parent.Name,
                        "Object Name": t.Name,
                        "Object Type": "Calculated Table",
                    }
                )
            else:
                rows.append(
                    {
                        "Parent Name": t.Parent.Name,
                        "Object Name": t.Name,
                        "Object Type": str(t.ObjectType),
                    }
                )
            for c in t.Columns:
                if str(c.Type) != "RowNumber":
                    if str(c.Type) == "Calculated":
                        rows.append(
                            {
                                "Parent Name": c.Parent.Name,
                                "Object Name": c.Name,
                                "Object Type": "Calculated Column",
                            }
                        )
                    else:
                        rows.append(
                            {
                                "Parent Name": c.Parent.Name,
                                "Object Name": c.Name,
                                "Object Type": str(c.ObjectType),
                            }
                        )
            for m in t.Measures:
                rows.append(
                    {
                        "Parent Name": m.Parent.Name,
                        "Object Name": m.Name,
                        "Object Type": str(m.ObjectType),
                    }
                )
            for h in t.Hierarchies:
                rows.append(
                    {
                        "Parent Name": h.Parent.Name,
                        "Object Name": h.Name,
                        "Object Type": str(h.ObjectType),
                    }
                )
                for lev in h.Levels:
                    rows.append(
                        {
                            "Parent Name": lev.Parent.Name,
                            "Object Name": lev.Name,
                            "Object Type": str(lev.ObjectType),
                        }
                    )
            for p in t.Partitions:
                rows.append(
                    {
                        "Parent Name": p.Parent.Name,
                        "Object Name": p.Name,
                        "Object Type": str(p.ObjectType),
                    }
                )
        for r in tom.model.Relationships:
            rName = create_relationship_name(
                r.FromTable.Name, r.FromColumn.Name, r.ToTable.Name, r.ToColumn.Name
            )
            rows.append(
                {
                    "Parent Name": r.Parent.Name,
                    "Object Name": rName,
                    "Object Type": str(r.ObjectType),
                }
            )
        for role in tom.model.Roles:
            rows.append(
                {
                    "Parent Name": role.Parent.Name,
                    "Object Name": role.Name,
                    "Object Type": str(role.ObjectType),
                }
            )
            for rls in role.TablePermissions:
                rows.append(
                    {
                        "Parent Name": role.Name,
                        "Object Name": rls.Name,
                        "Object Type": str(rls.ObjectType),
                    }
                )
        for tr in tom.model.Cultures:
            rows.append(
                {
                    "Parent Name": tr.Parent.Name,
                    "Object Name": tr.Name,
                    "Object Type": str(tr.ObjectType),
                }
            )
        for per in tom.model.Perspectives:
            rows.append(
                {
                    "Parent Name": per.Parent.Name,
                    "Object Name": per.Name,
                    "Object Type": str(per.ObjectType),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def list_shortcuts(
    lakehouse: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
    path: Optional[str] = None,
) -> pd.DataFrame:
    """
    Shows all shortcuts which exist in a Fabric lakehouse and their properties.

    *** NOTE: This function has been moved to the lakehouse subpackage. Please repoint your code to use that location. ***

    Parameters
    ----------
    lakehouse : str, default=None
        The Fabric lakehouse name.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | uuid.UUID, default=None
        The name or ID of the Fabric workspace in which lakehouse resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    path: str, default=None
        The path within lakehouse where to look for shortcuts. If provied, must start with either "Files" or "Tables". Examples: Tables/FolderName/SubFolderName; Files/FolderName/SubFolderName.
        Defaults to None which will retun all shortcuts on the given lakehouse

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing all the shortcuts which exist in the specified lakehouse.
    """

    from sempy_labs.lakehouse._shortcuts import list_shortcuts

    print(
        f"{icons.warning} This function has been moved to the lakehouse subpackage. Please repoint your code to use that location."
    )

    return list_shortcuts(lakehouse=lakehouse, workspace=workspace, path=path)


@log
def list_reports_using_semantic_model(
    dataset: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Shows a list of all the reports which use a given semantic model. This is limited to the reports which are in the same workspace as the semantic model.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the reports which use a given semantic model.
    """

    # df = pd.DataFrame(
    #    columns=[
    #        "Report Name",
    #        "Report Id",
    #        "Report Workspace Name",
    #        "Report Workspace Id",
    #    ]
    # )

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    dfR = fabric.list_reports(workspace=workspace_id)
    dfR_filt = dfR[
        (dfR["Dataset Id"] == dataset_id)
        & (dfR["Dataset Workspace Id"] == workspace_id)
    ][["Name", "Id"]]
    dfR_filt.rename(columns={"Name": "Report Name", "Id": "Report Id"}, inplace=True)
    dfR_filt["Report Workspace Name"] = workspace_name
    dfR_filt["Report Workspace Id"] = workspace_id

    return dfR_filt

    # response = _base_api(request=f"metadata/relations/downstream/dataset/{dataset_id}?apiVersion=3")

    # response_json = response.json()

    # for i in response_json.get("artifacts", []):
    #    object_workspace_id = i.get("workspace", {}).get("objectId")
    #    object_type = i.get("typeName")

    #    if object_type == "Report":
    #        new_data = {
    #            "Report Name": i.get("displayName"),
    #            "Report Id": i.get("objectId"),
    #            "Report Workspace Name": fabric.resolve_workspace_name(
    #                object_workspace_id
    #            ),
    #            "Report Workspace Id": object_workspace_id,
    #        }
    #        df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)


@log
def list_report_semantic_model_objects(
    dataset: str | UUID, workspace: Optional[str | UUID] = None, extended: bool = False
) -> pd.DataFrame:
    """
    Shows a list of semantic model objects (i.e. columns, measures, hierarchies) used in all reports which feed data from
    a given semantic model.

    Note: As with all functions which rely on the ReportWrapper, this function requires the report(s) to be in the 'PBIR' format.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    extended: bool, default=False
        If True, adds an extra column called 'Valid Semantic Model Object' which identifies whether the semantic model object used
        in the report exists in the semantic model which feeds data to the report.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of semantic model objects (i.e. columns, measures, hierarchies) used in all reports which feed data from
    a given semantic model.
    """

    from sempy_labs.report import ReportWrapper
    from sempy_labs.tom import connect_semantic_model

    columns = {
        "Report Name": "string",
        "Report Workspace Name": "string",
        "Table Name": "string",
        "Object Name": "string",
        "Object Type": "string",
        "Report Source": "string",
        "Report Source Object": "string",
    }
    dfRO = _create_dataframe(columns=columns)

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    # Collect all reports which use the semantic model
    dfR = list_reports_using_semantic_model(dataset=dataset_id, workspace=workspace_id)

    if len(dfR) == 0:
        return dfRO

    for _, r in dfR.iterrows():
        report_name = r["Report Name"]
        report_workspace = r["Report Workspace Name"]

        rpt = ReportWrapper(report=report_name, workspace=report_workspace)
        # Collect all semantic model objects used in the report
        dfRSO = rpt.list_semantic_model_objects()
        dfRSO["Report Name"] = report_name
        dfRSO["Report Workspace Name"] = report_workspace
        colName = "Report Name"
        dfRSO.insert(0, colName, dfRSO.pop(colName))
        colName = "Report Workspace Name"
        dfRSO.insert(1, colName, dfRSO.pop(colName))

        dfRO = pd.concat([dfRO, dfRSO], ignore_index=True)

    # Collect all semantic model objects
    if extended:
        with connect_semantic_model(
            dataset=dataset_id, readonly=True, workspace=workspace_id
        ) as tom:
            for index, row in dfRO.iterrows():
                object_type = row["Object Type"]
                if object_type == "Measure":
                    dfRO.at[index, "Valid Semantic Model Object"] = any(
                        o.Name == row["Object Name"] for o in tom.all_measures()
                    )
                elif object_type == "Column":
                    dfRO.at[index, "Valid Semantic Model Object"] = any(
                        format_dax_object_name(c.Parent.Name, c.Name)
                        == format_dax_object_name(row["Table Name"], row["Object Name"])
                        for c in tom.all_columns()
                    )
                elif object_type == "Hierarchy":
                    dfRO.at[index, "Valid Semantic Model Object"] = any(
                        format_dax_object_name(h.Parent.Name, h.Name)
                        == format_dax_object_name(row["Table Name"], row["Object Name"])
                        for h in tom.all_hierarchies()
                    )

    return dfRO


@log
def list_semantic_model_object_report_usage(
    dataset: str | UUID,
    workspace: Optional[str | UUID] = None,
    include_dependencies: bool = False,
    extended: bool = False,
) -> pd.DataFrame:
    """
    Shows a list of semantic model objects and how many times they are referenced in all reports which rely on this semantic model.

    Requirement: Reports must be in the PBIR format.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    include_dependencies : bool, default=False
        If True, includes measure dependencies.
    extended: bool, default=False
        If True, adds columns 'Total Size', 'Data Size', 'Dictionary Size', 'Hierarchy Size' based on Vertipaq statistics.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of semantic model objects and how many times they are referenced in all reports which rely on this semantic model. By default, the dataframe
        is sorted descending by 'Report Usage Count'.
    """

    from sempy_labs._model_dependencies import get_model_calc_dependencies
    from sempy_labs._helper_functions import format_dax_object_name

    workspace_id = resolve_workspace_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    fabric.refresh_tom_cache(workspace=workspace)

    dfR = list_report_semantic_model_objects(dataset=dataset_id, workspace=workspace_id)
    usage_column_name = "Report Usage Count"

    if not include_dependencies:
        final_df = (
            dfR.groupby(["Table Name", "Object Name", "Object Type"])
            .size()
            .reset_index(name=usage_column_name)
        )
    else:
        df = pd.DataFrame(columns=["Table Name", "Object Name", "Object Type"])
        dep = get_model_calc_dependencies(dataset=dataset_id, workspace=workspace_id)

        for i, r in dfR.iterrows():
            object_type = r["Object Type"]
            table_name = r["Table Name"]
            object_name = r["Object Name"]
            new_data = {
                "Table Name": table_name,
                "Object Name": object_name,
                "Object Type": object_type,
            }
            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)
            df["Object Type"] = df["Object Type"].replace(
                "Attribute Hierarchy", "Column"
            )
            if object_type in ["Measure", "Calc Column", "Calc Table", "Hierarchy"]:
                df_filt = dep[dep["Object Name"] == object_name][
                    ["Referenced Table", "Referenced Object", "Referenced Object Type"]
                ]
                df_filt.rename(
                    columns={
                        "Referenced Table": "Table Name",
                        "Referenced Object": "Object Name",
                        "Referenced Object Type": "Object Type",
                    },
                    inplace=True,
                )

                df = pd.concat([df, df_filt], ignore_index=True)

        final_df = (
            df.groupby(["Table Name", "Object Name", "Object Type"])
            .size()
            .reset_index(name=usage_column_name)
        )

    if extended:
        final_df["Object"] = format_dax_object_name(
            final_df["Table Name"], final_df["Object Name"]
        )
        dfC = fabric.list_columns(
            dataset=dataset_id, workspace=workspace_id, extended=True
        )
        dfC["Object"] = format_dax_object_name(dfC["Table Name"], dfC["Column Name"])
        final_df = pd.merge(
            final_df,
            dfC[
                [
                    "Object",
                    "Total Size",
                    "Data Size",
                    "Dictionary Size",
                    "Hierarchy Size",
                ]
            ],
            on="Object",
            how="left",
        )

        ext_int_cols = ["Total Size", "Data Size", "Dictionary Size", "Hierarchy Size"]
        final_df[ext_int_cols] = final_df[ext_int_cols].fillna(0).astype(int)
        final_df.drop("Object", axis=1, inplace=True)

    int_cols = [usage_column_name]
    final_df[int_cols] = final_df[int_cols].astype(int)

    final_df = final_df[final_df["Object Type"] != "Table"].sort_values(
        by=usage_column_name, ascending=False
    )

    final_df.reset_index(drop=True, inplace=True)

    return final_df


@log
def list_server_properties(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Lists the `properties <https://learn.microsoft.com/dotnet/api/microsoft.analysisservices.serverproperty?view=analysisservices-dotnet>`_ of the Analysis Services instance.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of the server properties.
    """

    tom_server = fabric.create_tom_server(
        dataset=None, readonly=True, workspace=workspace
    )

    rows = [
        {
            "Name": sp.Name,
            "Value": sp.Value,
            "Default Value": sp.DefaultValue,
            "Is Read Only": sp.IsReadOnly,
            "Requires Restart": sp.RequiresRestart,
            "Units": sp.Units,
            "Category": sp.Category,
        }
        for sp in tom_server.ServerProperties
    ]

    tom_server.Dispose()
    df = pd.DataFrame(rows)

    column_map = {
        "Is Read Only": "bool",
        "Requires Restart": "bool",
    }

    _update_dataframe_datatypes(dataframe=df, column_map=column_map)

    return df


@log
def list_semantic_model_errors(
    dataset: str | UUID, workspace: Optional[str | UUID]
) -> pd.DataFrame:
    """
    Shows a list of a semantic model's errors and their error messages (if they exist).

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of the errors and error messages for a given semantic model.
    """

    from sempy_labs.tom import connect_semantic_model

    df = pd.DataFrame(
        columns=["Object Type", "Table Name", "Object Name", "Error Message"]
    )

    error_rows = []

    with connect_semantic_model(
        dataset=dataset, workspace=workspace, readonly=True
    ) as tom:
        # Define mappings of TOM objects to object types and attributes
        error_checks = [
            ("Column", tom.all_columns, lambda o: o.ErrorMessage),
            ("Partition", tom.all_partitions, lambda o: o.ErrorMessage),
            (
                "Partition - Data Coverage Expression",
                tom.all_partitions,
                lambda o: (
                    o.DataCoverageDefinition.ErrorMessage
                    if o.DataCoverageDefinition
                    else ""
                ),
            ),
            ("Row Level Security", tom.all_rls, lambda o: o.ErrorMessage),
            ("Calculation Item", tom.all_calculation_items, lambda o: o.ErrorMessage),
            ("Measure", tom.all_measures, lambda o: o.ErrorMessage),
            (
                "Measure - Detail Rows Expression",
                tom.all_measures,
                lambda o: (
                    o.DetailRowsDefinition.ErrorMessage
                    if o.DetailRowsDefinition
                    else ""
                ),
            ),
            (
                "Measure - Format String Expression",
                tom.all_measures,
                lambda o: (
                    o.FormatStringDefinition.ErrorMessage
                    if o.FormatStringDefinition
                    else ""
                ),
            ),
            (
                "Calculation Group - Multiple or Empty Selection Expression",
                tom.all_calculation_groups,
                lambda o: (
                    o.CalculationGroup.MultipleOrEmptySelectionExpression.ErrorMessage
                    if o.CalculationGroup.MultipleOrEmptySelectionExpression
                    else ""
                ),
            ),
            (
                "Calculation Group - No Selection Expression",
                tom.all_calculation_groups,
                lambda o: (
                    o.CalculationGroup.NoSelectionExpression.ErrorMessage
                    if o.CalculationGroup.NoSelectionExpression
                    else ""
                ),
            ),
            ("Function", tom.all_functions, lambda o: o.ErrorMessage),
        ]

        # Iterate over all error checks
        for object_type, getter, error_extractor in error_checks:
            for obj in getter():
                error_message = error_extractor(obj)
                if error_message:  # Only add rows if there's an error message
                    error_rows.append(
                        {
                            "Object Type": object_type,
                            "Table Name": obj.Parent.Name,
                            "Object Name": obj.Name,
                            "Error Message": error_message,
                        }
                    )

    if error_rows:
        df = pd.DataFrame(error_rows)

    return df


@log
def list_synonyms(dataset: str | UUID, workspace: Optional[str] = None):

    from sempy_labs.tom import connect_semantic_model

    columns = {
        "Culture Name": "string",
        "Table Name": "string",
        "Object Name": "string",
        "Object Type": "string",
        "Synonym": "string",
        "Type": "string",
        "State": "string",
        "Source": "string",
        "Weight": "float_fillna",
        "Last Modified": "string",
    }

    df = _create_dataframe(columns=columns)

    rows = []
    with connect_semantic_model(
        dataset=dataset, workspace=workspace, readonly=True
    ) as tom:
        for c in tom.model.Cultures:
            if c.LinguisticMetadata is not None:
                lm = json.loads(c.LinguisticMetadata.Content)
                if "Entities" in lm:
                    for _, v in lm.get("Entities", []).items():
                        binding = v.get("Definition", {}).get("Binding", {})

                        t_name = binding.get("ConceptualEntity")
                        object_name = binding.get("ConceptualProperty")

                        if object_name is None:
                            object_type = "Table"
                            object_name = t_name
                        elif any(
                            m.Name == object_name and m.Parent.Name == t_name
                            for m in tom.all_measures()
                        ):
                            object_type = "Measure"
                        elif any(
                            m.Name == object_name and m.Parent.Name == t_name
                            for m in tom.all_columns()
                        ):
                            object_type = "Column"
                        elif any(
                            m.Name == object_name and m.Parent.Name == t_name
                            for m in tom.all_hierarchies()
                        ):
                            object_type = "Hierarchy"

                        merged_terms = defaultdict(dict)
                        for t in v.get("Terms", []):
                            for term, properties in t.items():
                                merged_terms[term].update(properties)

                        for term, props in merged_terms.items():
                            new_data = {
                                "Culture Name": lm.get("Language"),
                                "Table Name": t_name,
                                "Object Name": object_name,
                                "Object Type": object_type,
                                "Synonym": term,
                                "Type": props.get("Type"),
                                "State": props.get("State"),
                                "Source": props.get("Source", {}).get("Agent"),
                                "Weight": props.get("Weight"),
                                "Last Modified": props.get("LastModified"),
                            }

                            # Skip concatenation if new_data is empty or invalid
                            if any(new_data.values()):
                                rows.append(new_data)

    if rows:
        df = pd.DataFrame(rows)
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df



import pandas as pd
from typing import Optional
from sempy_labs._helper_functions import (
    _base_api,
    _create_dataframe,
    delete_item,
    resolve_workspace_id,
)
from uuid import UUID
from sempy._utils._log import log


@log
def list_operations_agents(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Returns a list of Operations Agents from the specified workspace.

    This is a wrapper function for the following API: `Items - List Operations Agents <https://learn.microsoft.com/rest/api/fabric/operationsagent/items/list-operations-agents>`_.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the Operations Agents within a workspace.
    """
    workspace_id = resolve_workspace_id(workspace)

    columns = {
        "Operations Agent Name": "str",
        "Operations Agent Id": "str",
        "Description": "str",
        "State": "str",
    }
    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request=f"v1/workspaces/{workspace_id}/OperationsAgents",
        client="fabric_sp",
        uses_pagination=True,
    )

    rows = []

    for r in responses:
        for v in r.get("value", []):
            row = {
                "Operations Agent Name": v.get("displayName"),
                "Operations Agent Id": v.get("id"),
                "Description": v.get("description"),
                "State": v.get("properties", {}).get("state"),
            }
            rows.append(row)

    return df


@log
def delete_operations_agent(
    operations_agent: str | UUID, workspace: Optional[str | UUID] = None
):
    """
    Deletes an Operations Agent from the specified workspace.

    This is a wrapper function for the following API: `Items - Delete Operations Agent <https://learn.microsoft.com/rest/api/fabric/operationsagent/items/delete-operations-agent>`_.

    Parameters
    ----------
    operations_agent : str | uuid.UUID
        The Operations Agent name or ID to delete.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    delete_item(item=operations_agent, type="OperationsAgent", workspace=workspace)



from uuid import UUID
import pandas as pd
from typing import Optional, List
from tqdm.auto import tqdm
from sempy_labs._helper_functions import (
    resolve_item_name_and_id,
    resolve_workspace_name_and_id,
    _base_api,
    _create_dataframe,
)
from IPython.display import HTML, display
import sempy_labs._icons as icons
import time
from sempy_labs.report._generate_embed_token import generate_embed_token
from sempy._utils._log import log


def embed_report_save_in_edit_mode(embed_url, access_token: str) -> pd.DataFrame:
    html = f"""
    <div id="reportContainer" style="height:800px;width:100%;display:none;"></div>

    <script src="https://cdn.jsdelivr.net/npm/powerbi-client@2.23.1/dist/powerbi.min.js"></script>
    <script>
        var models = window['powerbi-client'].models;

        // Embed configuration for the report
        var embedConfig = {{
            type: 'report',
            tokenType: models.TokenType.Embed,
            accessToken: '{access_token}',
            embedUrl: '{embed_url}',
            permissions: models.Permissions.ReadWrite,
            viewMode: models.ViewMode.Edit
        }};

        // Get the container element where the report will be embedded (hidden from user)
        var reportContainer = document.getElementById('reportContainer');

        // Embed the report (will be invisible)
        var report = powerbi.embed(reportContainer, embedConfig);

        // Listen for the 'rendered' event to ensure the report is fully loaded
        report.on('rendered', function() {{
            console.log("Report rendered successfully in background.");

            // Trigger save once the report is rendered
            report.save().then(function() {{
                console.log("Report saved successfully in background!");
            }}).catch(function(error) {{
                console.error("Error saving the report:", error);
            }});
        }});

        // Error handling for embed
        report.on('error', function(event) {{
            console.error("Error embedding the report:", event.detail);
        }});
    </script>
    """
    display(HTML(html))


@log
def upgrade_to_pbir(
    report: Optional[str | UUID | List[str | UUID]] = None,
    workspace: Optional[str | UUID | List[str | UUID]] = None,
) -> pd.DataFrame:
    """
    Upgrades a Power BI report to the new `PBIR <https://powerbi.microsoft.com/blog/power-bi-enhanced-report-format-pbir-in-power-bi-desktop-developer-mode-preview>`_ format.

    Parameters
    ----------
    report : str | uuid.UUID | typing.List[str | uuid.UUID], default=None
        Name or ID of the Power BI report. Also accepts a list of report name/IDs.
        If set to None, upgrades all eligible reports in the specified workspace(s).
    workspace : str | uuid.UUID | typing.List[str | uuid.UUID], default=None
        The name or ID of the Fabric workspace(s).
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the format of all reports in the specified workspace(s) after conducting the upgrade.
    """

    if isinstance(workspace, (str, UUID)) or workspace is None:
        workspace = [workspace]

    workspaces = {}
    for w in workspace:
        workspace_name, workspace_id = resolve_workspace_name_and_id(w)
        workspaces[workspace_id] = workspace_name

    columns = {
        "Workspace Name": "str",
        "Workspace Id": "str",
        "Report Name": "str",
        "Report Id": "str",
        "Format": "str",
    }

    df = _create_dataframe(columns=columns)

    rows = []
    for workspace_id, workspace_name in workspaces.items():
        url = f"/v1.0/myorg/groups/{workspace_id}/reports"
        response = _base_api(request=url, client="fabric_sp")

        eligible_for_upgrade = {}
        # Collect reports eligible for upgrade within the workspace
        for rpt in response.json().get("value", []):
            rpt_id = rpt.get("id")
            rpt_name = rpt.get("name")
            rpt_format = rpt.get("format")
            embed_url = rpt.get("embedUrl")
            dataset_id = rpt.get("datasetId")
            if (
                rpt_format == "PBIRLegacy"
                and dataset_id is not None
                and embed_url is not None
            ):
                eligible_for_upgrade[rpt_id] = (embed_url, dataset_id, rpt_name)

        # Determine which reports to process
        if report is None:
            report_ids_to_process = list(eligible_for_upgrade.keys())
        else:
            if isinstance(report, str):
                report = [report]

            report_ids_to_process = []

            for r in report:
                rpt_name, rpt_id = resolve_item_name_and_id(
                    item=r, type="Report", workspace=workspace_id
                )

                if rpt_id in eligible_for_upgrade:
                    report_ids_to_process.append(rpt_id)
                else:
                    print(
                        f"{icons.warning} The {rpt_name} report in the '{workspace_name}' workspace is not eligible for upgrade."
                    )

        # Run upgrade for selected reports
        total_reports = len(report_ids_to_process)
        for idx, rpt_id in enumerate(
            bar := tqdm(report_ids_to_process, desc="Upgrading reports", unit="report"),
            start=1,
        ):
            embed_url, dataset_id, rpt_name = eligible_for_upgrade[rpt_id]
            bar.set_description(
                f"Upgrading the '{rpt_name}' report within the '{workspace_name}' workspace ({idx}/{total_reports})"
            )

            access_token = generate_embed_token(
                dataset_ids=[dataset_id],
                report_ids=[rpt_id],
            )

            embed_report_save_in_edit_mode(embed_url, access_token)

            row = check_upgrade_status(
                report_id=rpt_id,
                report_name=rpt_name,
                workspace_id=workspace_id,
                workspace_name=workspace_name,
            )

            rows.append(row)

        # Double check if there are any reports that were not upgraded yet and check their status again (to account for any potential delay in the upgrade process)
        not_upgraded_yet = [
            row["Report Id"] for row in rows if row.get("Format") == "PBIRLegacy"
        ]
        if not_upgraded_yet:
            for r in rows:
                rpt_id = r.get("Report Id")
                if rpt_id in not_upgraded_yet:
                    rpt_name = r.get("Report Name")
                    ws_id = r.get("Workspace Id")
                    ws_name = r.get("Workspace Name")
                    row = check_upgrade_status(
                        report_id=rpt_id,
                        report_name=rpt_name,
                        workspace_id=ws_id,
                        workspace_name=ws_name,
                        verbose=True,
                        time_limit=120,
                    )  # check for up to 2 minutes
                    if row.get("Format") == "PBIR":
                        # Update the row in the list of rows
                        for idx, existing_row in enumerate(rows):
                            if existing_row.get("Report Id") == rpt_id:
                                rows[idx] = row
                                break

    if rows:
        df = pd.DataFrame(rows)

    return df


# Define the time limit (2 minute)
TIME_BETWEEN_REQUESTS = 2  # seconds


# Function to check the upgrade status
def check_upgrade_status(
    report_id, report_name, workspace_id, workspace_name, verbose=False, time_limit=30
):
    start_time = time.time()
    while time.time() - start_time < time_limit:
        response = _base_api(
            request=f"/v1.0/myorg/groups/{workspace_id}/reports/{report_id}",
            client="fabric_sp",
        )
        format = response.json().get("format")
        if format == "PBIR":
            break

        # Wait for 2 seconds before the next request
        time.sleep(TIME_BETWEEN_REQUESTS)

    if format != "PBIR" and verbose:
        print(
            f"{icons.yellow_dot} The '{report_name}' report within the '{workspace_name}' workspace has not yet been upgraded to PBIR format."
        )

    return {
        "Workspace Name": workspace_name,
        "Workspace Id": workspace_id,
        "Report Name": report_name,
        "Report Id": report_id,
        "Format": format,
    }



from sempy_labs.directlake._generate_shared_expression import generate_shared_expression
from sempy_labs._helper_functions import (
    resolve_dataset_name_and_id,
    resolve_workspace_name_and_id,
    resolve_item_name_and_id,
    resolve_lakehouse_name_and_id,
)
from sempy._utils._log import log
from sempy_labs.tom import connect_semantic_model
from typing import Optional, List, Literal
import sempy_labs._icons as icons
from uuid import UUID
from sempy_labs.directlake._sources import (
    _get_direct_lake_expressions,
    _extract_expression_list,
)


@log
def update_direct_lake_model_lakehouse_connection(
    dataset: str | UUID,
    workspace: Optional[str | UUID] = None,
    lakehouse: Optional[str] = None,
    lakehouse_workspace: Optional[str | UUID] = None,
):
    """
    Remaps a Direct Lake semantic model's SQL Endpoint connection to a new lakehouse.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid. UUID, default=None
        The Fabric workspace name or ID in which the semantic model exists.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse used by the Direct Lake semantic model.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    lakehouse_workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    update_direct_lake_model_connection(
        dataset=dataset,
        workspace=workspace,
        source=lakehouse,
        source_type="Lakehouse",
        source_workspace=lakehouse_workspace,
    )


@log
def update_direct_lake_model_connection(
    dataset: str | UUID,
    workspace: Optional[str | UUID] = None,
    source: Optional[str | UUID] = None,
    source_type: Literal["Lakehouse", "Warehouse"] = "Lakehouse",
    source_workspace: Optional[str | UUID] = None,
    use_sql_endpoint: bool = True,
    tables: Optional[str | List[str]] = None,
):
    """
    Remaps a Direct Lake semantic model's SQL Endpoint connection to a new lakehouse/warehouse.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID in which the semantic model exists.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    source : str | uuid.UUID, default=None
        The name of the Fabric lakehouse/warehouse used by the Direct Lake semantic model.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    source_type : typing.Literal["Lakehouse", "Warehouse"], default="Lakehouse"
        The type of source for the Direct Lake semantic model. Valid options: "Lakehouse", "Warehouse".
    source_workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse/warehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    use_sql_endpoint : bool, default=True
        If True, the SQL Endpoint will be used for the connection.
        If False, Direct Lake over OneLake will be used.
    tables : str | List[str], default=None
        The name(s) of the table(s) to update in the Direct Lake semantic model.
        If None, all tables will be updated (if there is only one expression).
        If multiple tables are specified, they must be provided as a list.
    """
    if use_sql_endpoint:
        icons.sll_tags.append("UpdateDLConnection_SQL")
    else:
        icons.sll_tags.append("UpdateDLConnection_DLOL")

    if isinstance(tables, str):
        tables = [tables]

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    source_type = source_type.capitalize()

    if source_type not in ["Lakehouse", "Warehouse"]:
        raise ValueError(
            f"{icons.red_dot} The 'source_type' must be either 'Lakehouse' or 'Warehouse'."
        )

    if source_workspace is None:
        source_workspace = workspace_name

    if source_type == "Lakehouse":
        (source_name, source_id) = resolve_lakehouse_name_and_id(
            lakehouse=source, workspace=source_workspace
        )
    else:
        (source_name, source_id) = resolve_item_name_and_id(
            item=source, type=source_type, workspace=source_workspace
        )

    shared_expression = generate_shared_expression(
        item_name=source_name,
        item_type=source_type,
        workspace=source_workspace,
        use_sql_endpoint=use_sql_endpoint,
    )

    expression_dict = _get_direct_lake_expressions(dataset=dataset, workspace=workspace)
    expressions = list(expression_dict.keys())

    with connect_semantic_model(
        dataset=dataset_id, readonly=False, workspace=workspace_id
    ) as tom:

        if not tom.is_direct_lake():
            raise ValueError(
                f"{icons.red_dot} The '{dataset_name}' semantic model within the '{workspace_name}' workspace is not in Direct Lake. This function is only applicable to Direct Lake semantic models."
            )

        # Update the single connection expression
        if len(expressions) > 1 and not tables:
            print(
                f"{icons.info} Multiple expressions found in the model. Please specify the tables to update using the 'tables parameter."
            )
            return
        elif len(expressions) == 1 and not tables:
            expr = expressions[0]
            tom.model.Expressions[expr].Expression = shared_expression

            # Fix for converting to DL on SQL. Without this fix, the measures and relationships will disappear from the model.
            if use_sql_endpoint:
                expr_name = 'DatabaseQuery'
                tom.model.Expressions[expr].Name = expr_name
                tom.set_annotation(object=tom.model, name='PBI_QueryOrder', value=expr_name)

            print(
                f"{icons.green_dot} The expression in the '{dataset_name}' semantic model within the '{workspace_name}' workspace has been updated to point to the '{source}' {source_type.lower()} in the '{source_workspace}' workspace."
            )
        else:
            import sempy

            sempy.fabric._client._utils._init_analysis_services()
            import Microsoft.AnalysisServices.Tabular as TOM

            expr_list = _extract_expression_list(shared_expression)

            expr_name = next(
                (name for name, exp in expression_dict.items() if exp == expr_list),
                None,
            )

            # If the expression does not already exist, create it
            def generate_unique_name(existing_names):
                i = 1
                while True:
                    candidate = f"DatabaseQuery{i}"
                    if candidate not in existing_names:
                        return candidate
                    i += 1

            if not expr_name:
                expr_name = generate_unique_name(expressions)
                tom.add_expression(name=expr_name, expression=shared_expression)

            all_tables = [t.Name for t in tom.model.Tables]
            for t_name in tables:
                if t_name not in all_tables:
                    raise ValueError(
                        f"{icons.red_dot} The table '{t_name}' does not exist in the '{dataset_name}' semantic model within the '{workspace_name}' workspace."
                    )
                p = next(p for p in tom.model.Tables[t_name].Partitions)
                if p.Mode != TOM.ModeType.DirectLake:
                    raise ValueError(
                        f"{icons.red_dot} The table '{t_name}' in the '{dataset_name}' semantic model within the '{workspace_name}' workspace is not in Direct Lake mode. This function is only applicable to Direct Lake tables."
                    )

                p.Source.ExpressionSource = tom.model.Expressions[expr_name]



from ._secrets import (
    get_secret,
    set_secret,
    delete_secret,
    list_secrets,
    recover_deleted_secret,
    purge_deleted_secret,
    backup_secret,
    restore_secret,
    update_secret,
    list_deleted_secrets,
    list_secret_versions,
)


__all__ = [
    "get_secret",
    "set_secret",
    "delete_secret",
    "list_secrets",
    "recover_deleted_secret",
    "purge_deleted_secret",
    "backup_secret",
    "restore_secret",
    "update_secret",
    "list_deleted_secrets",
    "list_secret_versions",
]



from typing import Any
import pandas as pd
from typing import Tuple
from sempy_labs._helper_functions import (
    _base_api,
    _create_dataframe,
)
import sempy_labs._icons as icons
from sempy._utils._log import log


def resolve_secret_name_and_version(
    key_vault_uri: str, secret_id: str
) -> Tuple[str, str]:

    if not key_vault_uri.endswith("/"):
        key_vault_uri = f"{key_vault_uri}/"

    secret_name = secret_id.split(f"{key_vault_uri}secrets/")[1]
    if "/" in secret_name:
        return (secret_name.split("/")[0], secret_name.split("/")[1])
    else:
        return (secret_name, None)


@log
def get_secret(key_vault_uri: str, secret: str) -> Any:
    """
    Retrieves the latest version of a secret from the specified Azure Key Vault.

    This is a wrapper function for the following API: `Get Secret - Get Secret <https://learn.microsoft.com/rest/api/keyvault/secrets/get-secret/get-secret>`_.

    Parameters
    ----------
    key_vault_uri : str
        Azure Key Vault URI.
    secret : str
        Name of the secret in the Key Vault.

    Returns
    -------
    Any
        The value of the latest version of the secret.
    """

    response = _base_api(
        request=f"{key_vault_uri}/secrets/{secret}/versions", client="keyvault"
    )
    url = response.json().get("value")[-1].get("id")
    response = _base_api(request=f"{url}", client="keyvault")
    return response.json().get("value")


@log
def set_secret(key_vault_uri: str, name: str, value: Any):
    """
    Sets a secret in the specified Azure Key Vault.

    This is a wrapper function for the following API: `Set Secret - Set Secret <https://learn.microsoft.com/rest/api/keyvault/secrets/set-secret/set-secret>`_.

    Parameters
    ----------
    key_vault_uri : str
        Azure Key Vault URI.
    name : str
        Name of the secret to be set in the Key Vault.
    value : Any
        Value of the secret to be set in the Key Vault.
    """

    payload = {"value": value}
    _base_api(
        request=f"{key_vault_uri}/secrets/{name}",
        client="keyvault",
        method="put",
        payload=payload,
    )
    print(
        f"{icons.green_dot} The '{name}' secret has been successfully set within the '{key_vault_uri}' Key Vault."
    )


@log
def list_secrets(key_vault_uri: str) -> pd.DataFrame:
    """
    Lists all secrets in the specified Azure Key Vault.

    This is a wrapper function for the following API: `Get Secrets - Get Secrets <https://learn.microsoft.com/rest/api/keyvault/secrets/get-secrets/get-secrets>`_.

    Parameters
    ----------
    key_vault_uri : str
        Azure Key Vault URI.

    Returns
    -------
    pandas.DataFrame
        A pandas DataFrame containing details of all secrets in the Key Vault.
    """

    columns = {
        "Secret Id": "str",
        "Secret Name": "str",
        "Enabled": "bool",
        "Created Date": "datetime",
        "Updated Date": "datetime",
        "Recovery Level": "str",
        "Recoverable Days": "int",
    }

    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request=f"{key_vault_uri}/secrets", client="keyvault", uses_pagination=True
    )

    rows = []
    for r in responses:
        secret_id = r.get("id")
        secret_name = resolve_secret_name_and_version(
            key_vault_uri=key_vault_uri, secret_id=secret_id
        )[0]
        rows.append(
            {
                "Secret Id": secret_id,
                "Secret Name": secret_name,
                "Enabled": r.get("attributes", {}).get("enabled"),
                "Created Date": r.get("attributes", {}).get("created"),
                "Updated Date": r.get("attributes", {}).get("updated"),
                "Recovery Level": r.get("attributes", {}).get("recoveryLevel"),
                "Recoverable Days": r.get("attributes", {}).get("recoverableDays"),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        df["Created Date"] = pd.to_datetime(df["Created Date"], unit="s")
        df["Updated Date"] = pd.to_datetime(df["Updated Date"], unit="s")
        df["Recoverable Days"] = df["Recoverable Days"].astype(int)

    return df


@log
def list_secret_versions(key_vault_uri: str, secret: str) -> pd.DataFrame:
    """
    Lists all versions of a specific secret in the specified Azure Key Vault.

    This is a wrapper function for the following API: `Get Secret Versions - Get Secret Versions <https://learn.microsoft.com/rest/api/keyvault/secrets/get-secret-versions/get-secret-versions>`_.

    Parameters
    ----------
    key_vault_uri : str
        Azure Key Vault URI.
    secret : str
        Name of the secret.

    Returns
    -------
    pandas.DataFrame
        A pandas DataFrame containing details of all versions of the secret in the Key Vault.
    """

    columns = {
        "Secret Id": "str",
        "Secret Name": "str",
        "Secret Version": "str",
        "Enabled": "bool",
        "Created Date": "datetime",
        "Updated Date": "datetime",
    }

    df = _create_dataframe(columns=columns)

    url = f"{key_vault_uri}/secrets/{secret}/versions"
    responses = _base_api(request=url, client="keyvault", uses_pagination=True)

    rows = []
    for r in responses:
        secret_id = r.get("id")
        (secret_name, secret_version) = resolve_secret_name_and_version(
            key_vault_uri=key_vault_uri, secret_id=secret_id
        )
        rows.append(
            {
                "Secret Id": secret_id,
                "Secret Name": secret_name,
                "Secret Version": secret_version,
                "Enabled": r.get("attributes", {}).get("enabled"),
                "Created Date": r.get("attributes", {}).get("created"),
                "Updated Date": r.get("attributes", {}).get("updated"),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        df["Created Date"] = pd.to_datetime(df["Created Date"], unit="s")
        df["Updated Date"] = pd.to_datetime(df["Updated Date"], unit="s")

    return df


@log
def list_deleted_secrets(key_vault_uri: str) -> pd.DataFrame:
    """
    Lists all deleted secrets in the specified Azure Key Vault.

    This is a wrapper function for the following API: `Get Deleted Secrets - GetDeleted Secrets <https://learn.microsoft.com/rest/api/keyvault/secrets/get-deleted-secrets/get-deleted-secrets>`_.

    Parameters
    ----------
    key_vault_uri : str
        Azure Key Vault URI.

    Returns
    -------
    pandas.DataFrame
        A pandas DataFrame containing details of all secrets in the Key Vault.
    """

    columns = {
        "Recovery Id": "str",
        "Deleted Date": "datetime",
        "Scheduled Purge Date": "datetime",
        "Content Type": "str",
        "Secret Id": "str",
        "Secret Name": "str",
        "Enabled": "bool",
        "Created Date": "datetime",
        "Updated Date": "datetime",
        "Recovery Level": "str",
    }

    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request=f"{key_vault_uri}/deletedSecrets",
        client="keyvault",
        uses_pagination=True,
    )

    rows = []
    for r in responses:
        secret_id = r.get("id")
        secret_name = resolve_secret_name_and_version(
            key_vault_uri=key_vault_uri, secret_id=secret_id
        )[0]
        rows.append(
            {
                "Recovery Id": r.get("recoveryId"),
                "Deleted Date": r.get("deletedDate"),
                "Scheduled Purge Date": r.get("scheduledPurgeDate"),
                "Content Type": r.get("contentType"),
                "Secret Id": secret_id,
                "Secret Name": secret_name,
                "Enabled": r.get("attributes", {}).get("enabled"),
                "Created Date": r.get("attributes", {}).get("created"),
                "Updated Date": r.get("attributes", {}).get("updated"),
                "Recovery Level": r.get("attributes", {}).get("recoveryLevel"),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        df["Scheduled Purge Date"] = pd.to_datetime(
            df["Scheduled Purge Date"], unit="s"
        )
        df["Deleted Date"] = pd.to_datetime(df["Deleted Date"], unit="s")
        df["Created Date"] = pd.to_datetime(df["Created Date"], unit="s")
        df["Updated Date"] = pd.to_datetime(df["Updated Date"], unit="s")

    return df


@log
def recover_deleted_secret(key_vault_uri: str, secret: str):
    """
    Recovers a deleted secret in the specified Azure Key Vault.

    This is a wrapper function for the following API: `Recover Deleted Secret - Recover Deleted Secret <https://learn.microsoft.com/rest/api/keyvault/secrets/recover-deleted-secret/recover-deleted-secret>`_.

    Parameters
    ----------
    key_vault_uri : str
        Azure Key Vault URI.
    secret : str
        Name of the deleted secret to be recovered in the Key Vault.
    """

    _base_api(
        request=f"{key_vault_uri}/deletedsecrets/{secret}/recover",
        client="keyvault",
        method="post",
    )
    print(
        f"{icons.green_dot} The '{secret}' secret within the '{key_vault_uri}' Key Vault has been successfully recovered."
    )


@log
def purge_deleted_secret(key_vault_uri: str, secret: str):
    """
    Permanently deletes the specified secret.
    The purge deleted secret operation removes the secret permanently, without the possibility of recovery. This operation can only be enabled on a soft-delete enabled vault. This operation requires the secrets/purge permission.

    This is a wrapper function for the following API: `Purge Deleted Secret - Purge Deleted Secret <https://learn.microsoft.com/rest/api/keyvault/secrets/purge-deleted-secret/purge-deleted-secret>`_.

    Parameters
    ----------
    key_vault_uri : str
        Azure Key Vault URI.
    secret : str
        Name of the deleted secret to be recovered in the Key Vault.
    """

    _base_api(
        request=f"{key_vault_uri}/deletedsecrets/{secret}",
        client="keyvault",
        method="delete",
        status_codes=204,
    )
    print(
        f"{icons.green_dot} The '{secret}' secret within the '{key_vault_uri}' Key Vault has been successfully purged."
    )


@log
def delete_secret(key_vault_uri: str, secret: str):
    """
    Deletes a secret in the specified Azure Key Vault.

    This is a wrapper function for the following API: `Delete Secret - Delete Secret <https://learn.microsoft.com/rest/api/keyvault/secrets/delete-secret/delete-secret>`_.

    Parameters
    ----------
    key_vault_uri : str
        Azure Key Vault URI.
    secret : str
        Name of the secret to be deleted in the Key Vault.
    """

    _base_api(
        request=f"{key_vault_uri}/secrets/{secret}",
        client="keyvault",
        method="delete",
    )
    print(
        f"{icons.green_dot} The '{secret}' secret has been successfully deleted within the '{key_vault_uri}' Key Vault."
    )


@log
def backup_secret(key_vault_uri: str, secret: str) -> bytes:
    """
    Backs up a secret from the specified Azure Key Vault.

    This is a wrapper function for the following API: `Backup Secret - Backup Secret <https://learn.microsoft.com/rest/api/keyvault/secrets/backup-secret/backup-secret>`_.

    Parameters
    ----------
    key_vault_uri : str
        Azure Key Vault URI.
    secret : str
        Name of the secret to be backed up in the Key Vault.

    Returns
    -------
    bytes
        The backup blob of the secret.
    """

    response = _base_api(
        request=f"{key_vault_uri}/secrets/{secret}/backup", client="keyvault"
    )
    return response.json().get("value")


@log
def restore_secret(key_vault_uri: str, backup_blob: bytes) -> dict:
    """
    Restores a backed up secret to the specified Azure Key Vault.

    This is a wrapper function for the following API: `Restore Secret - Restore Secret <https://learn.microsoft.com/rest/api/keyvault/secrets/restore-secret/restore-secret>`_.

    Parameters
    ----------
    key_vault_uri : str
        Azure Key Vault URI.
    backup_blob : bytes
        The backup blob of the secret to be restored.

    Returns
    -------
    dict
        The restored secret details.
    """

    payload = {"value": backup_blob}
    response = _base_api(
        request=f"{key_vault_uri}/secrets/restore",
        client="keyvault",
        method="post",
        payload=payload,
    )

    print(
        f"{icons.green_dot} The secret has been successfully restored within the '{key_vault_uri}' Key Vault."
    )

    return response.json()


@log
def update_secret(
    key_vault_uri: str, secret: str, version: str, enabled: bool = None
) -> dict:
    """
    Updates the attributes of a secret in the specified Azure Key Vault.

    This is a wrapper function for the following API: `Update Secret - Update Secret <https://learn.microsoft.com/rest/api/keyvault/secrets/update-secret/update-secret>`_.

    Parameters
    ----------
    key_vault_uri : str
        Azure Key Vault URI.
    secret : str
        Name of the secret to be updated in the Key Vault.
    version : str
        Version of the secret to be updated in the Key Vault.
    enabled : bool, optional
        Specifies whether the secret is enabled or disabled.

    Returns
    -------
    dict
        The updated secret attributes.
    """

    payload = {"attributes": {"enabled": enabled}}

    response = _base_api(
        request=f"{key_vault_uri}/secrets/{secret}/{version}",
        client="keyvault",
        method="patch",
        payload=payload,
    )
    print(
        f"{icons.green_dot} The '{secret}' secret has been successfully updated within the '{key_vault_uri}' Key Vault."
    )

    return response.json()



"""
Microsoft Fabric & Power BI REST API Fuzzy Search Tool

This module provides fuzzy/ambiguous search functionality for Fabric and Power BI TOC JSON.
It uses the `rapidfuzz` library for efficient fuzzy string matching.

Usage:
    from fabric_api_search import search_api, fetch_toc

    toc = fetch_toc()  # Fetches both Fabric and Power BI
    results = search_api(toc, "copy fabric job")

    # Search only Fabric APIs
    results = search_api(toc, "copy job", source="fabric")

    # Search only Power BI APIs
    results = search_api(toc, "dataset", source="powerbi")

    # Or run directly:
    python fabric_api_search.py "copy fabric job"
    python fabric_api_search.py "dataset" --source powerbi

Requirements:
    pip install rapidfuzz requests
"""

import argparse
import json
from typing import Dict, List, Literal

import requests
from rapidfuzz import fuzz, process

# Constants
FABRIC_BASE_URL = "https://learn.microsoft.com/en-us/rest/api/fabric"
POWERBI_BASE_URL = "https://learn.microsoft.com/en-us/rest/api/power-bi"

TOC_SOURCES = {
    "fabric": {
        "toc_url": f"{FABRIC_BASE_URL}/toc.json",
        "base_url": FABRIC_BASE_URL,
        "name": "Microsoft Fabric"
    },
    "powerbi": {
        "toc_url": f"{POWERBI_BASE_URL}/toc.json",
        "base_url": POWERBI_BASE_URL,
        "name": "Power BI"
    }
}

SourceType = Literal["fabric", "powerbi", None]


def fetch_toc(source: SourceType = None) -> Dict:
    """
    Fetch the Table of Contents JSON from Microsoft Learn.

    Args:
        source: Which API source to fetch:
            - "fabric": Only Fabric APIs
            - "powerbi": Only Power BI APIs
            - None: Both Fabric and Power BI APIs (default)

    Returns:
        Dict: The combined TOC data with source information
    """
    sources_to_fetch = []

    if source is None:
        sources_to_fetch = ["fabric", "powerbi"]
    elif source in TOC_SOURCES:
        sources_to_fetch = [source]
    else:
        raise ValueError(f"Invalid source: {source}. Must be 'fabric', 'powerbi', or None")

    combined_items = []

    for src in sources_to_fetch:
        src_config = TOC_SOURCES[src]
        print(f"📥 Fetching {src_config['name']} TOC...")

        try:
            response = requests.get(src_config["toc_url"], timeout=30)
            response.raise_for_status()
            toc_data = response.json()

            # Add source metadata to each top-level item
            items = toc_data.get("items", [])
            for item in items:
                item["_source"] = src
                item["_base_url"] = src_config["base_url"]

            combined_items.extend(items)
            print(f"   ✅ Loaded {len(items)} top-level categories from {src_config['name']}")

        except requests.RequestException as e:
            print(f"   ⚠️ Failed to fetch {src_config['name']} TOC: {e}")

    return {"items": combined_items}


def _build_full_url(href: str, base_url: str) -> str:
    """
    Build the full documentation URL from a relative href.

    Args:
        href: The relative API path (e.g., "core/connections/list-supported-connection-types")
        base_url: The base URL for the API source

    Returns:
        str: Full URL (e.g., "https://learn.microsoft.com/en-us/rest/api/fabric/core/connections/list-supported-connection-types")
    """
    if href.startswith("http"):
        return href
    return f"{base_url}/{href}"


def _flatten_toc_items(items: List[Dict], parent_path: str = "",
                       source: str = None, base_url: str = None) -> List[Dict]:
    """
    Recursively flatten TOC items into a list of leaf nodes with href.

    Args:
        items: List of TOC items
        parent_path: Accumulated path from parent items
        source: The source identifier (fabric/powerbi)
        base_url: The base URL for building full URLs

    Returns:
        List of flattened items with href, full URL, and searchable text
    """
    flattened = []

    for item in items:
        title = item.get("toc_title", "")
        current_path = f"{parent_path} > {title}" if parent_path else title

        # Get source info from item or inherit from parent
        item_source = item.get("_source", source)
        item_base_url = item.get("_base_url", base_url)

        # If item has href, it's a leaf node (API endpoint)
        if "href" in item:
            href = item["href"]
            full_url = _build_full_url(href, item_base_url) if item_base_url else href

            # Convert href path to readable text
            href_text = href.replace("/", " ").replace("-", " ").replace("(", " ").replace(")", " ")

            # Add source name to searchable text for filtering
            source_name = TOC_SOURCES.get(item_source, {}).get("name", "") if item_source else ""

            flattened.append({
                "href": href,
                "url": full_url,
                "toc_title": title,
                "path": current_path,
                "source": item_source,
                "searchable_text": f"{title} {href_text} {current_path} {source_name}".lower()
            })

        # Recursively process children
        if "children" in item:
            flattened.extend(_flatten_toc_items(
                item["children"],
                current_path,
                item_source,
                item_base_url
            ))

    return flattened


def search_api(toc: Dict, query: str, limit: int = 20, score_threshold: int = 50,
               source: SourceType = None) -> List[Dict]:
    """
    Perform fuzzy search on the Fabric/Power BI TOC JSON to find matching API entries.

    This function uses rapidfuzz for fuzzy string matching, supporting ambiguous
    queries like "copy fabric job" to find entries like "Create Copy Job".

    Args:
        toc: The raw TOC JSON dictionary (as returned by fetch_toc())
        query: The search query string (e.g., "copy fabric job")
        limit: Maximum number of results to return (default: 20)
        score_threshold: Minimum fuzzy match score (0-100) to include in results (default: 50)
        source: Filter results by source:
            - "fabric": Only Fabric APIs
            - "powerbi": Only Power BI APIs
            - None: Both (default)

    Returns:
        List of matching API entries, each containing:
        - href: The relative API path
        - url: The full documentation URL
        - toc_title: The API title
        - path: The hierarchical path
        - source: The API source (fabric/powerbi)
        - score: The fuzzy match score (0-100)

    Example:
        >>> toc = fetch_toc()
        >>> results = search_api(toc, "copy fabric job")
        >>> for r in results:
        ...     print(f"{r['toc_title']}: {r['url']}")
        Create Copy Job: https://learn.microsoft.com/en-us/rest/api/fabric/copyjob/items/create-copy-job
        ...
    """
    # Get items from TOC
    items = toc.get("items", [])
    if not items:
        return []

    # Flatten the TOC structure
    flattened = _flatten_toc_items(items)

    # Filter by source if specified
    if source:
        flattened = [item for item in flattened if item.get("source") == source]

    if not flattened:
        return []

    # Normalize query for matching
    query_normalized = query.lower().strip()

    # Use rapidfuzz to find matches
    choices = {i: item["searchable_text"] for i, item in enumerate(flattened)}

    # Use token_set_ratio for best matching with different word orders and partial matches
    matches = process.extract(
        query_normalized,
        choices,
        scorer=fuzz.token_set_ratio,
        limit=limit * 2
    )

    # Also try partial_ratio for substring matches
    partial_matches = process.extract(
        query_normalized,
        choices,
        scorer=fuzz.partial_ratio,
        limit=limit * 2
    )

    # Combine and deduplicate matches, taking the best score for each item
    score_map: Dict[int, float] = {}
    for match_text, score, idx in matches:
        if idx not in score_map or score > score_map[idx]:
            score_map[idx] = score

    for match_text, score, idx in partial_matches:
        combined_score = (score_map.get(idx, 0) * 0.6) + (score * 0.4)
        if idx not in score_map or combined_score > score_map[idx]:
            score_map[idx] = combined_score

    # Filter by threshold and sort by score
    results = []
    for idx, score in sorted(score_map.items(), key=lambda x: x[1], reverse=True):
        if score >= score_threshold:
            item = flattened[idx]
            results.append({
                "href": item["href"],
                "url": item["url"],
                "toc_title": item["toc_title"],
                "path": item["path"],
                "source": item["source"],
                "score": round(score, 2)
            })

    return results[:limit]


def search_api_simple(toc: Dict, query: str, limit: int = 20,
                      source: SourceType = None) -> List[Dict]:
    """
    Simplified version that returns only essential fields.

    Args:
        toc: The raw TOC JSON dictionary
        query: The search query string
        limit: Maximum number of results to return
        source: Filter by source (fabric/powerbi/None)

    Returns:
        List of dicts with 'href', 'url', 'toc_title', and 'source' keys
    """
    results = search_api(toc, query, limit=limit, source=source)
    return [{
        "href": r["href"],
        "url": r["url"],
        "toc_title": r["toc_title"],
        "source": r["source"]
    } for r in results]


def main():
    """Command-line interface for the search tool."""
    parser = argparse.ArgumentParser(
        description="Fuzzy search for Microsoft Fabric and Power BI REST APIs"
    )
    parser.add_argument(
        "query",
        type=str,
        help="Search query (e.g., 'copy fabric job', 'dataset refresh')"
    )
    parser.add_argument(
        "--source",
        type=str,
        choices=["fabric", "powerbi"],
        default=None,
        help="Filter by API source: 'fabric', 'powerbi', or omit for both"
    )
    parser.add_argument(
        "--limit",
        type=int,
        default=20,
        help="Maximum number of results (default: 20)"
    )
    parser.add_argument(
        "--threshold",
        type=int,
        default=50,
        help="Minimum match score 0-100 (default: 50)"
    )
    parser.add_argument(
        "--json",
        action="store_true",
        help="Output results as JSON"
    )
    parser.add_argument(
        "--simple",
        action="store_true",
        help="Output simplified format"
    )

    args = parser.parse_args()

    source_desc = args.source.upper() if args.source else "Fabric + Power BI"
    print(f"🔍 Searching for: '{args.query}' in {source_desc}")
    print("=" * 80)

    # Fetch TOC (fetch both, filter later for flexibility)
    toc = fetch_toc(source=args.source)

    print()

    # Perform search
    if args.simple:
        results = search_api_simple(toc, args.query, limit=args.limit, source=args.source)
    else:
        results = search_api(toc, args.query, limit=args.limit,
                             score_threshold=args.threshold, source=args.source)

    # Output results
    if args.json:
        print(json.dumps(results, indent=2))
    else:
        if not results:
            print("No results found.")
        else:
            print(f"Found {len(results)} results:\n")
            for i, result in enumerate(results, 1):
                source_tag = f"[{result['source'].upper()}]" if result.get('source') else ""
                if args.simple:
                    print(f"{i}. {source_tag} {result['toc_title']}")
                    print(f"   URL: {result['url']}")
                else:
                    print(f"{i}. {source_tag} {result['toc_title']} (score: {result['score']})")
                    print(f"   URL: {result['url']}")
                    print(f"   Path: {result['path']}")
                print()


if __name__ == "__main__":
    main()



from ._create_pqt_file import create_pqt_file
from ._migrate_calctables_to_lakehouse import (
    migrate_calc_tables_to_lakehouse,
    migrate_field_parameters,
)
from ._migrate_calctables_to_semantic_model import (
    migrate_calc_tables_to_semantic_model,
)
from ._migrate_model_objects_to_semantic_model import (
    migrate_model_objects_to_semantic_model,
)
from ._migrate_tables_columns_to_semantic_model import (
    migrate_tables_columns_to_semantic_model,
)
from ._migration_validation import (
    migration_validation,
)
from ._refresh_calc_tables import (
    refresh_calc_tables,
)
from ._direct_lake_to_import import (
    migrate_direct_lake_to_import,
)

__all__ = [
    "create_pqt_file",
    "migrate_calc_tables_to_lakehouse",
    "migrate_field_parameters",
    "migrate_calc_tables_to_semantic_model",
    "migrate_model_objects_to_semantic_model",
    "migrate_tables_columns_to_semantic_model",
    "migration_validation",
    "refresh_calc_tables",
    "migrate_direct_lake_to_import",
]



from uuid import UUID
from typing import Optional
import pandas as pd
from sempy_labs._helper_functions import (
    _base_api,
    _create_dataframe,
    resolve_workspace_id,
    resolve_item_id,
)
from sempy._utils._log import log


@log
def list_reports_base(
    report: Optional[str | UUID] = None, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:

    columns = {
        "Report Name": "string",
        "Report Id": "string",
        "Format": "string",
        "Web Url": "string",
        "Embed Url": "string",
        "Is From PBIX": "bool",
        "Is Owned By Me": "bool",
        "Dataset Id": "string",
        "Dataset Workspace Id": "string",
        "Users": "list",
        "Subscriptions": "list",
    }

    workspace_id = resolve_workspace_id(workspace)
    url = f"/v1.0/myorg/groups/{workspace_id}/reports"

    if report is not None:
        report_id = resolve_item_id(item=report, type="Report", workspace=workspace_id)
        url += f"/{report_id}"

    response_json = _base_api(url).json()

    fields = {
        "Report Name": "name",
        "Report Id": "id",
        "Format": "format",
        "Web Url": "webUrl",
        "Embed Url": "embedUrl",
        "Is From PBIX": "isFromPbix",
        "Is Owned By Me": "isOwnedByMe",
        "Dataset Id": "datasetId",
        "Dataset Workspace Id": "datasetWorkspaceId",
        "Users": "users",
        "Subscriptions": "subscriptions",
    }

    items = [response_json] if report is not None else response_json.get("value", [])

    rows = [
        {
            col: item.get(src, [] if columns[col] == "list" else None)
            for col, src in fields.items()
        }
        for item in items
    ]

    df = (
        pd.DataFrame(rows, columns=columns.keys())
        if rows
        else _create_dataframe(columns=columns)
    )

    return df


@log
def list_reports(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Returns a list of reports from the specified workspace.

    This is a wrapper function for the following API: `Reports - Get Reports In Group <https://learn.microsoft.com/rest/api/power-bi/reports/get-reports-in-group>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the list of reports.
    """

    return list_reports_base(workspace=workspace)


@log
def get_report(
    report: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Returns the properties of a specific report from the specified workspace.

    This is a wrapper function for the following API: `Reports - Get Report In Group <https://learn.microsoft.com/rest/api/power-bi/reports/get-report-in-group>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    report : str | uuid.UUID
        Name or ID of the Power BI report.
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the properties of the report.
    """

    return list_reports_base(report=report, workspace=workspace)



from uuid import UUID
from typing import Optional
from sempy_labs._helper_functions import (
    get_model_id,
    resolve_item_name_and_id,
    resolve_workspace_name_and_id,
    _base_api,
)
from sempy._utils._log import log
import sempy_labs._icons as icons


@log
def enable_query_caching(
    dataset: str | UUID,
    workspace: Optional[str | UUID] = None,
    enable: bool = True,
):
    """
    Enables or disables `query caching <http://aka.ms/queryCaching>`_ for a semantic model.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    enable : bool, default=True
        Set to True to enable query caching, or False to disable it.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (item_name, item_id) = resolve_item_name_and_id(
        item=dataset, type="SemanticModel", workspace=workspace_id
    )
    model_id = get_model_id(item_id=item_id)
    if model_id is None:
        raise ValueError(
            f"Failed to retrieve model ID for semantic model '{item_name}'"
        )

    caching_map = {
        True: 2,
        False: 1,
    }

    payload = {"queryCachingState": caching_map.get(enable)}

    _base_api(
        request=f"metadata/models/{model_id}/caching",
        method="internal",
        payload=payload,
        status_codes=204,
    )

    print(
        f"{icons.green_dot} Query caching has been {'enabled' if enable else 'disabled'} for the '{item_name}' semantic model within the '{workspace_name}' workspace."
    )
